from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common import Icons
from biscuit.common.ui import Bubble, Frame

if typing.TYPE_CHECKING:
    from . import Statusbar


class SBubble(Bubble):
    def get_pos(self) -> str:
        return (
            f"+{int(self.master.winfo_rootx() + (self.master.winfo_width() - self.winfo_width())/2)}"
            + f"+{self.master.winfo_rooty() - self.master.winfo_height() - 15}"
        )


class SButton(Frame):
    def __init__(
        self,
        master: Statusbar,
        text: str = None,
        icon: Icons = None,
        callback=None,
        description: str = None,
        highlighted: bool = False,
        padx: int = 5,
        pady: int = 1,
        icon2: Icons = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, padx=padx, pady=pady, *args, **kwargs)
        self.callback = callback or (lambda *_: None)
        self.text = text
        self.icon = icon
        self.highlighted = highlighted
        self.icon2 = icon2
        self.toggled = False

        self.bg, self.fg, self.hbg, self.hfg = (
            self.base.theme.layout.statusbar.button_highlighted.values()
            if self.highlighted
            else self.base.theme.layout.statusbar.button.values()
        )
        self.config(bg=self.bg)

        self.bubble = SBubble(self, text=description)
        if icon:
            self.icon_label = tk.Label(
                self,
                text=self.icon,
                anchor=tk.CENTER,
                bg=self.bg,
                fg=self.fg,
                font=("codicon", 12),
            )
            self.icon_label.pack(side=tk.LEFT, fill=tk.Y, expand=True)

        if text:
            self.text_label = tk.Label(
                self,
                text=self.text,
                anchor=tk.CENTER,
                bg=self.bg,
                fg=self.fg,
                pady=2,
                font=self.base.settings.uifont,
            )
            self.text_label.pack(side=tk.LEFT, fill=tk.Y, expand=True)

        self.config_bindings()
        self.visible = False

    def config_bindings(self) -> None:
        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)

        self.bind("<Button-1>", self.on_click)
        if self.text:
            self.text_label.bind("<Button-1>", self.on_click)
        if self.icon:
            self.icon_label.bind("<Button-1>", self.on_click)

    def on_enter(self, *_) -> None:
        # warning: bad code
        self.bubble.show()
        if self.highlighted:
            self.config(bg=self.hbg)
        if self.text:
            self.text_label.config(fg=self.hfg)
            if self.highlighted:
                self.text_label.config(bg=self.hbg)

        if self.icon:
            self.icon_label.config(fg=self.hfg)
            if self.highlighted:
                self.icon_label.config(bg=self.hbg)

    def on_leave(self, *_) -> None:
        # warning: bad code
        self.bubble.hide()
        if self.highlighted:
            self.config(bg=self.bg)
        if self.text:
            self.text_label.config(fg=self.fg)
            if self.highlighted:
                self.text_label.config(bg=self.bg)
        if self.icon:
            self.icon_label.config(fg=self.fg)
            if self.highlighted:
                self.icon_label.config(bg=self.bg)

    def on_click(self, *_) -> None:
        self.callback()
        self.bubble.hide()

        if self.icon2:
            self.toggled = not self.toggled
            self.icon_label.config(text=self.icon2 if self.toggled else self.icon)

    def change_text(self, text: str) -> None:
        self.text_label.config(text=text)

    def change_description(self, text: str) -> None:
        self.bubble.change_text(text)

    def change_icon(self, icon: Icons) -> None:
        self.icon_label.config(text=icon)

    def change_callback(self, callback: typing.Callable) -> None:
        self.callback = callback

    def set_pack_data(self, **kwargs) -> None:
        self.pack_data = kwargs

    def get_pack_data(self) -> None:
        return self.pack_data

    def show(self) -> None:
        if not self.visible:
            self.lift()
            self.visible = True
            self.pack(**self.get_pack_data(), fill=tk.Y)

    def hide(self) -> None:
        if self.visible:
            self.visible = False
            self.pack_forget()



from ..language.languages import Languages

comment_prefix_map = {
    Languages.PYTHON: "#",
    Languages.JULIA: "#",
    Languages.VB_NET: "'",
    Languages.JAVASCRIPT: "//",
    Languages.TYPESCRIPT: "//",
    Languages.C: "//",
    Languages.CPP: "//",
    Languages.CSHARP: "//",
    Languages.D: "//",
    Languages.RUBY: "#",
    Languages.PHP: "//",
    Languages.RUST: "//",
    Languages.SWIFT: "//",
    Languages.GO: "//",
    Languages.KOTLIN: "//",
    Languages.JAVA: "//",
    Languages.SCALA: "//",
    Languages.HASKELL: "--",
    Languages.SQL: "--",
    Languages.BASH: "#",
    Languages.POWERSHELL: "#",
    Languages.JSON: "//",
    Languages.YAML: "#",
    Languages.TOML: "#",
    Languages.LUA: "--",
    Languages.DART: "//",
}


def get_comment_prefix(language: str) -> str:
    """Get the comment prefix for a language."""

    return comment_prefix_map.get(language, None)


def register_comment_prefix(language: str, prefix: str):
    """Register a comment prefix for a language."""

    comment_prefix_map[language] = prefix



import tkinter as tk

from biscuit.common.ui import Frame, Shortcut


class Shortcuts(Frame):
    """Visual representation of the shortcuts"""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.row = 0

    def add_shortcut(self, name: str, value: tuple[str]) -> None:
        name = tk.Label(
            self,
            text=name,
            font=self.base.settings.uifont,
            anchor=tk.E,
            **self.base.theme.editors.labels
        )
        shortcut = Shortcut(self, keys=value, **self.base.theme.editors)

        name.grid(row=self.row, column=0, sticky=tk.EW, pady=5, padx=5)
        shortcut.grid(row=self.row, column=1, sticky=tk.EW, pady=5, padx=5)

        self.row += 1



from __future__ import annotations

import typing

from .loader import WorkspaceLoader
from .workspace import Workspace

if typing.TYPE_CHECKING:
    from biscuit.app import App


class WorkspaceManager:
    def __init__(self, base: App):
        self.base = base
        self.workspace = None

        self.loader = WorkspaceLoader(self)

    def set_workspace(self, workspace: Workspace):
        self.workspace = workspace
        self.base.workspace_opened()

    def add_dir(self, dir: str):
        if not self.workspace:
            self.set_workspace(Workspace(self, "workspace.toml", [dir]))
            self.set_workspace(Workspace(self, "workspace.toml", [dir]))
            return

        self.workspace.add_dir(dir)
        self.base.workspace_changed(dir)

    def load(self, path: str):
        self.loader.load(path)

    def save(self, path: str):
        self.loader.save(path)

    def close(self):
        self.workspace = None
        self.base.workspace_closed()



import os

from .app import App
from .common import check_python_installation


def get_app_instance(
    exec_dir: str = os.path.abspath(__file__),
    open_path: str = None,
) -> App:
    """Get an instance of the application.

    Args:
        exec_dir (str, optional): The directory of the executable. Defaults to os.path.dirname(os.path.abspath(__file__)).
        open_path (str, optional): The path to open. Defaults to None.

    Returns:
        App: An instance of the application."""

    check_python_installation()
    return App(exec_dir, dir=open_path)


def main(args: list[str] = []):
    """Main entry point for the application.

    Args:
        args (list[str], optional): Command line arguments. Defaults to []."""

    dir = None
    if not args:
        args = [os.path.abspath(__file__)]
    elif len(args) >= 2:
        dir = args[1]

    app = get_app_instance(args[0], dir)
    app.run()


def start_app():
    """Start the application."""

    import sys

    main(sys.argv)


if __name__ == "__main__":
    start_app()




__author__ = "tomlin7"

import random
import tkinter as tk

from biscuit.common.ui import Button

from .gamebase import BaseGame

GRID_WIDTH = 50
GRID_HEIGHT = 50
CELL_SIZE = 20
INITIAL_ALIVE_PROBABILITY = 0.2


class GameOfLife(BaseGame):
    name = "Game of Life"

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, None, None, *args, **kwargs)

        self.start_stop_button = Button(
            self,
            text="Start",
            command=self.toggle_start_stop,
            padx=20,
            pady=20,
            font=("Fixedsys", 15),
        )
        self.canvas = tk.Canvas(
            self,
            width=GRID_WIDTH * CELL_SIZE,
            height=GRID_HEIGHT * CELL_SIZE,
            borderwidth=0,
            highlightthickness=0,
            **self.base.theme.editors
        )

        self._grid = [[0] * GRID_WIDTH for _ in range(GRID_HEIGHT)]
        # self.randomize_grid()
        self.draw_gridlines()
        self.is_running = False

        self.start_stop_button.pack()
        self.canvas.pack()

        self.canvas.bind("<B1-Motion>", self.add_cell)
        self.canvas.bind("<Button-1>", self.add_cell)
        self.canvas.bind("<B3-Motion>", self.remove_cell)
        self.canvas.bind("<Button-3>", self.remove_cell)

        self.base.notifications.info(
            "INSTRUCTIONS: Left click to add cells, Right click to remove cells!"
        )

    def randomize_grid(self) -> None:
        for row in range(GRID_HEIGHT):
            for col in range(GRID_WIDTH):
                if random.random() < INITIAL_ALIVE_PROBABILITY:
                    self._grid[row][col] = 1

    def draw_gridlines(self) -> None:
        for col in range(GRID_WIDTH + 1):
            x = col * CELL_SIZE
            self.canvas.create_line(
                x,
                0,
                x,
                GRID_HEIGHT * CELL_SIZE,
                fill=self.base.theme.border,
                tags="grid",
            )
        for row in range(GRID_HEIGHT + 1):
            y = row * CELL_SIZE
            self.canvas.create_line(
                0,
                y,
                GRID_WIDTH * CELL_SIZE,
                y,
                fill=self.base.theme.border,
                tags="grid",
            )

    def toggle_start_stop(self, *_) -> None:
        if self.is_running:
            self.is_running = False
            self.start_stop_button.config(text="Start")
        else:
            self.is_running = True
            self.start_stop_button.config(text="Stop")
            self.update()

    def add_cell(self, event) -> None:
        col = event.x // CELL_SIZE
        row = event.y // CELL_SIZE
        self._grid[row][col] = not self._grid[row][col]
        self.draw_cell(row, col)

    def remove_cell(self, event) -> None:
        col = event.x // CELL_SIZE
        row = event.y // CELL_SIZE
        self._grid[row][col] = 0
        self.erase_cell(row, col)

    def update(self) -> None:
        if not self.is_running:
            return

        new_grid = [[0] * GRID_WIDTH for _ in range(GRID_HEIGHT)]

        # rules
        for row in range(GRID_HEIGHT):
            for col in range(GRID_WIDTH):
                alive_neighbors = self.count_alive_neighbors(row, col)

                if self._grid[row][col] == 1:
                    if alive_neighbors < 2 or alive_neighbors > 3:
                        new_grid[row][col] = 0
                    else:
                        new_grid[row][col] = 1
                else:
                    if alive_neighbors == 3:
                        new_grid[row][col] = 1

        self._grid = new_grid
        self.draw_grid()
        self.after(100, self.update)

    def count_alive_neighbors(self, row, col) -> None:
        count = 0
        for i in range(-1, 2):
            for j in range(-1, 2):
                neighbor_row = (row + i + GRID_HEIGHT) % GRID_HEIGHT
                neighbor_col = (col + j + GRID_WIDTH) % GRID_WIDTH
                count += self._grid[neighbor_row][neighbor_col]

        count -= self._grid[row][col]
        return count

    def draw_grid(self) -> None:
        self.canvas.delete("cell")
        for row in range(GRID_HEIGHT):
            for col in range(GRID_WIDTH):
                if self._grid[row][col] == 1:
                    self.draw_cell(row, col)

    def draw_cell(self, row, col) -> None:
        x1 = col * CELL_SIZE
        y1 = row * CELL_SIZE
        x2 = x1 + CELL_SIZE
        y2 = y1 + CELL_SIZE
        self.canvas.create_rectangle(
            x1, y1, x2, y2, fill=self.base.theme.biscuit, outline="", tags="cell"
        )

    def erase_cell(self, row, col) -> None:
        x1 = col * CELL_SIZE
        y1 = row * CELL_SIZE
        x2 = x1 + CELL_SIZE
        y2 = y1 + CELL_SIZE
        self.canvas.create_rectangle(
            x1,
            y1,
            x2,
            y2,
            fill=self.base.theme.editors.background,
            outline=self.base.theme.border,
            tags="cell",
        )



import os

from biscuit.common.icons import Icons
from biscuit.common.ui import Label

from ..terminalbase import TerminalBase


class Default(TerminalBase):
    """
    Default Terminal - Checks COMSPEC/SHELL environmental variables set in the host machine
    and opens that in terminal. Shows Not Detected in case variable is not set.
    """

    # get the correct shell command depending on platform
    shell = os.environ.get("COMSPEC") or os.environ.get("SHELL")
    name = os.path.splitext(os.path.basename(shell))[0] if shell else None
    icon = Icons.TERMINAL

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        if not self.shell:
            Label(
                self,
                text="No shells detected for the host os, report an issue otherwise.",
            ).grid()
            self.name = "Not Detected"
            self.icon = "error"
            return

        self.start_service()



import tkinter as tk

from biscuit.common.ui import Button, Frame, IconLabelButton, WrappingLabel


class DebugPlaceholder(Frame):
    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=15, pady=10, **self.base.theme.views.sidebar.item)
        self.columnconfigure(0, weight=1)

        WrappingLabel(
            self,
            text="No debugger configuration found for active editor",
            font=self.base.settings.uifont,
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content
        ).grid(row=0, sticky=tk.EW)



from __future__ import annotations

import json
import queue
import threading
import tkinter as tk
import typing

import requests

from biscuit.common.ui import ScrollableFrame

from ..sidebar_item import SideBarViewItem
from .extension import ExtensionGUI
from .placeholder import ExtensionsPlaceholder

if typing.TYPE_CHECKING:
    from biscuit.extensions import ExtensionManager


class ExtensionsList(ScrollableFrame):
    items: list[ExtensionGUI]

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.views.sidebar.background)


class Results(SideBarViewItem):
    """View that displays the available extensions.
    # TODO search filter for extensions

    The Results view displays the available extensions.
    - The user can install, uninstall, search for extensions.
    """

    manager: ExtensionManager
    fetch_queue: queue.Queue
    fetching: threading.Event

    def __init__(self, master, *args, **kwargs) -> None:
        self.__actions__ = ()
        self.title = "Available"

        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.views.sidebar.item)

        self.placeholder = ExtensionsPlaceholder(self)
        self.extension_list = ExtensionsList(self.content)
        self.extension_list.pack(fill=tk.BOTH, expand=True)
        self.master = master

        self.filter_installed = False

        # self.watcher = ExtensionsWatcher(self)
        # self.watcher.watch()

        # TODO list installed extensions separately

    def late_setup(self) -> None:
        self.manager = self.base.extensions_manager
        self.fetch_queue = self.manager.fetch_queue
        self.fetching = self.manager.fetching

    def show_content(self) -> None:
        self.extension_list.pack(in_=self.content, fill=tk.BOTH, expand=True)
        self.placeholder.pack_forget()

    def show_placeholder(self) -> None:
        self.extension_list.pack_forget()
        self.placeholder.pack(in_=self.content, fill=tk.BOTH, expand=True)

    def refresh(self, *_) -> None:
        if self.base.testing:
            return

        self.clear()
        self.update_idletasks()
        self.after(5, self.manager.run_fetch_extensions())

    def gui_refresh_loop(self) -> None:
        if not self.fetch_queue.empty():
            name, data = self.fetch_queue.get()
            ext = ExtensionGUI(self, name, data)
            if not self.filter_installed or (self.filter_installed and ext.installed):
                self.extension_list.add(ext, fill=tk.X)

        self.after(5, self.gui_refresh_loop)

    def clear(self, *_) -> None:
        for widget in self.extension_list.items:
            widget.pack_forget()
            widget.destroy()
            self.content.update_idletasks()

        self.extension_list.items = []
        self.fetching.set()

    def set_selected(self, extension: ExtensionGUI) -> None:
        for widget in self.extension_list.items:
            if widget == extension:
                widget.select()
            else:
                widget.deselect()

        self.manager.open_extension(extension)

    def toggle_installed(self) -> None:
        self.filter_installed = not self.filter_installed
        self.refresh()

    def search(self) -> None:
        if self.base.testing:
            return

        self.clear()
        self.update_idletasks()
        self.manager.fetch_searched_extensions(self.master.searchbox.get())



from __future__ import annotations

from collections.abc import Mapping

from pygments.token import Token


class ThemeObject(Mapping):
    """
    Base theme object. If no colors are passed, uses the parent's color data.

    Parameters:
    parent, background, foreground, highlightbackground, highlightforeground,
    selectedbackground, selectedforeground
    """

    def __init__(
        self,
        master,
        background: str = None,
        foreground: str = None,
        highlightbackground: str = None,
        highlightforeground: str = None,
        selectedbackground: str = None,
        selectedforeground: str = None,
        **kwargs,
    ) -> None:
        self.master = master
        self.background = background or master.background
        self.foreground = foreground or master.foreground
        self.highlightbackground = highlightbackground or master.highlightbackground
        self.highlightforeground = highlightforeground or master.highlightforeground
        self.selectedbackground = (
            selectedbackground or highlightbackground or master.selectedbackground
        )
        self.selectedforeground = (
            selectedforeground or highlightforeground or master.selectedforeground
        )

        for key, value in kwargs.items():
            setattr(self, key, value)

    def values(self):
        """Returns a tuple of color values, order:
        Background, Foreground, HighlightBackground, HighlightForeground"""

        return (
            self.background,
            self.foreground,
            self.highlightbackground,
            self.highlightforeground,
        )

    def to_dict(self):
        return {
            "background": self.background,
            "foreground": self.foreground,
        }

    def update(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self

    def __getitem__(self, key):
        return self.to_dict()[key]

    def __iter__(self):
        return iter(self.to_dict())

    def __len__(self):
        return len(self.to_dict())

    def remove_bg_highlight(self):
        self.highlightbackground = self.background
        return self

    # TODO update child theme objects if parent's color values are altered after initialization
    # NOTE experimental code for updating child themeobjects
    # def update_child_colors(self, color_name, color_value):
    #     for attr_name in dir(self):
    #         attr = getattr(self, attr_name)
    #         if isinstance(attr, ThemeObject):
    #             setattr(attr, color_name, color_value)

    # def __setattr__(self, name, value):
    #     super().__setattr__(name, value)
    #     self.update_child_colors(name, value)

    # def __getattr__(self, name):
    #     return self.to_dict().get(name)


class HighlightableThemeObject(ThemeObject):
    def to_dict(self) -> dict:
        return {
            "background": self.background,
            "foreground": self.foreground,
            "activebackground": self.highlightbackground,
            "activeforeground": self.highlightforeground,
        }


class FrameThemeObject(ThemeObject):
    def to_dict(self) -> dict:
        return {"background": self.background}


class EditorsPane(FrameThemeObject):
    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.theme: Theme = master.theme

        self.bar = FrameThemeObject(self)
        self.bar.tab = HighlightableThemeObject(
            self.bar, highlightbackground=self.theme.secondary_background
        )
        self.bar.tab.icon = ThemeObject(self.bar.tab)
        self.bar.tab.close = HighlightableThemeObject(self.bar)


class PanelPane(FrameThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.bar = FrameThemeObject(self)
        self.bar.tab = HighlightableThemeObject(self.bar).remove_bg_highlight()


class ContentPane(FrameThemeObject):
    """
    ContentPane
    ├── EditorsPane
    └── Panel
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.theme: Theme = master.theme

        self.editors = EditorsPane(self)
        self.panel = PanelPane(self, *self.theme.secondary)


class DrawerPane(FrameThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.actionbar = FrameThemeObject(self)
        self.actionbar.slot = HighlightableThemeObject(
            self.actionbar
        ).remove_bg_highlight()
        self.actionbar.bubble = ThemeObject(self.actionbar)


class Layout(FrameThemeObject):
    def __init__(self, theme, *args, **kwargs) -> None:
        super().__init__(theme, *args, **kwargs)
        self.theme = theme

        self.menubar = FrameThemeObject(self)
        self.menubar.item = HighlightableThemeObject(self.menubar)
        self.menubar.title = ThemeObject(self)

        self.content = ContentPane(self)
        self.sidebar = DrawerPane(self)

        self.statusbar = FrameThemeObject(self)
        self.statusbar.button = HighlightableThemeObject(self.statusbar)
        self.statusbar.button_highlighted = HighlightableThemeObject(
            self.statusbar,
            self.master.biscuit,
            "white",
            self.master.biscuit_dark,
            "white",
        )


class SidebarViews(FrameThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.title = ThemeObject(self)
        self.itembar = FrameThemeObject(self)
        self.itembar.title = ThemeObject(self)
        self.item = FrameThemeObject(self)
        self.item.content = ThemeObject(self)
        self.item.button = HighlightableThemeObject(self)
        self.toggle_button = HighlightableThemeObject(self)


class PanelViews(FrameThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.button = HighlightableThemeObject(self)

        self.logs = ThemeObject(self)
        self.logs.time = "#008000"
        self.logs.caller = "#0000ff"
        self.logs.info = "#098677"
        self.logs.warning = "#a31515"
        self.logs.error = "#ab1515"

        self.terminal = FrameThemeObject(self)
        self.terminal.content = ThemeObject(self)
        self.terminal.tab = HighlightableThemeObject(self)


class Views(FrameThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.panel = PanelViews(self)
        self.sidebar = SidebarViews(self)


class Palette(FrameThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.searchbar = ThemeObject(self)
        self.item = HighlightableThemeObject(self)


class Menu(FrameThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.item = HighlightableThemeObject(self)


class Notifications(FrameThemeObject):
    def __init__(self, master: Theme, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.theme = master
        self.title = ThemeObject(self, fg=master.secondary_foreground)
        self.button = HighlightableThemeObject(self)
        self.text = ThemeObject(self)
        self.source = ThemeObject(
            self,
            master.primary_background,
            master.primary_foreground,
        )


class Editors(FrameThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        theme = self.master

        self.biscuit_labels = FrameThemeObject(self)  # welcome page
        self.button = HighlightableThemeObject(self)
        self.section = HighlightableThemeObject(self)  # settings
        self.labels = ThemeObject(self)

        self.bracket_colors = ("ffd700", "da70d6", "179fff")
        self.activebracket = "yellow"

        self.hyperlink = "#4583b6"

        self.selection = "#264f78"
        self.currentline = theme.border
        self.currentword = "#d5d5d5"
        self.found = "#dbe6c2"
        self.foundcurrent = "green"
        self.hovertag = "#d5d5d5"

        self.text = ThemeObject(self)
        self.minimap = FrameThemeObject(self)

        self.breadcrumbs = FrameThemeObject(self)
        self.breadcrumbs.item = HighlightableThemeObject(self).remove_bg_highlight()

        self.linenumbers = FrameThemeObject(self)
        self.linenumbers.number = ThemeObject(self.linenumbers)
        self.linenumbers.number.foreground = "#6e7681"
        self.linenumbers.number.highlightforeground = "#171184"
        self.linenumbers.breakpoint = HighlightableThemeObject(
            self.linenumbers, "#f5a199", "white", "#e51400", "white"
        )

        self.autocomplete = FrameThemeObject(self)
        self.autocomplete.item = ThemeObject(self.autocomplete)

        self.definitions = FrameThemeObject(self)
        self.definitions.item = HighlightableThemeObject(self.definitions)

        self.hover = FrameThemeObject(self)
        self.hover.text = ThemeObject(self.hover)

        self.diff = FrameThemeObject(self)
        self.diff.not_exist = "#d3d3d3"
        self.diff.removed = "#ffa3a3"
        self.diff.addition = "#dbe6c2"


class Utils(ThemeObject):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        theme: Theme = self.master

        self.button = HighlightableThemeObject(
            self, theme.biscuit, "#1e1e2e", theme.biscuit_dark
        )
        self.label = ThemeObject(self)
        self.linklabel = HighlightableThemeObject(
            self,
            theme.secondary_background,
            theme.biscuit,
            self.highlightbackground,
            theme.biscuit_light,
        )
        self.colorlabel = ThemeObject(self, theme.biscuit, "white", theme.biscuit_dark)
        self.tree = FrameThemeObject(self)
        self.tree.item = ThemeObject(self.tree)
        self.secondary_tree = FrameThemeObject(self, *theme.secondary)
        self.secondary_tree.item = ThemeObject(self.secondary_tree)

        self.bubble = ThemeObject(self)
        self.iconbutton = HighlightableThemeObject(self)
        self.iconbutton_hfg = HighlightableThemeObject(
            self, highlightbackground=self.background
        )
        self.iconbutton_hfg.highlightforeground = "white"
        self.iconlabelbutton = HighlightableThemeObject(
            self,
            theme.secondary_background,
            theme.secondary_foreground,
        )
        self.iconlabelbutton_hfg = HighlightableThemeObject(
            self,
            theme.secondary_background,
            theme.secondary_foreground,
            theme.secondary_background,
        )
        self.entry = ThemeObject(
            self,
            theme.secondary_background,
            theme.secondary_foreground,
            self.highlightbackground,
        )
        self.frame = FrameThemeObject(self)
        self.buttonsentry = ThemeObject(
            self,
            theme.secondary_background,
            theme.secondary_foreground,
            self.highlightbackground,
        )
        self.buttonsentry.button = HighlightableThemeObject(self.buttonsentry)
        self.bubble = ThemeObject(self)
        self.scrollbar = HighlightableThemeObject(self)


class Theme:
    """
    Following are the attributes available:
    layout, views, editors, palette, menu, notifications, utils
    """

    name = "default"

    biscuit = "#dc8c34"
    biscuit_light = "#ecb464"
    biscuit_dark = "#B56711"

    border = "#dfdfdf"
    disabled = "#8e8e90"

    primary_background = "#f8f8f8"
    primary_foreground = "#424242"
    primary_background_highlight = "#e1e1e1"
    primary_foreground_highlight = "black"

    secondary_background = "white"
    secondary_foreground = "black"
    secondary_background_highlight = "white"
    secondary_foreground_highlight = "black"

    # syntax
    Keyword = "#0000ff"
    Name = "#267f99"
    Function = "#795e26"
    String = "#b11515"
    Number = "#098658"
    Comment = {"foreground": "#098658", "font": {"slant": "italic"}}
    Punctuation = "#3b3b3b"

    def __init__(self) -> None:
        self.primary = primary = [
            self.primary_background,
            self.primary_foreground,
            self.primary_background_highlight,
            self.primary_foreground_highlight,
        ]
        self.secondary = secondary = [
            self.secondary_background,
            self.secondary_foreground,
            self.secondary_background_highlight,
            self.secondary_foreground_highlight,
        ]

        self.foreground = self.primary_foreground
        self.background = self.primary_background
        self.highlightbackground = self.primary_background_highlight
        self.highlightforeground = self.primary_foreground_highlight
        self.selectedbackground = self.primary_background_highlight
        self.selectedforeground = self.primary_foreground_highlight

        self.layout = Layout(self, *primary)
        self.views = Views(self, *primary)
        self.utils = Utils(self, *primary)
        self.menu = Menu(self, *secondary)
        self.editors = Editors(self, *secondary)
        self.palette = Palette(self, *secondary)
        self.findreplace = FrameThemeObject(self, *primary)
        self.findreplace.label = ThemeObject(self.findreplace)
        self.notifications = Notifications(self, *secondary)

        self.syntax = {
            Token.Keyword: self.Keyword,
            Token.Keyword.Constant: self.Keyword,
            Token.Keyword.Declaration: self.Keyword,
            Token.Keyword.Namespace: self.Keyword,
            Token.Keyword.Pseudo: self.Keyword,
            Token.Keyword.Reserved: self.Keyword,
            Token.Keyword.Type: self.Keyword,
            Token.Name: self.Name,
            Token.Name.Attribute: self.Name,
            Token.Name.Builtin: self.Name,
            Token.Name.Builtin.Pseudo: self.Name,
            Token.Name.Class: self.Name,
            Token.Name.Constant: self.Name,
            Token.Name.Decorator: self.Name,
            Token.Name.Entity: self.Name,
            Token.Name.Exception: self.Name,
            Token.Name.Function: self.Function,
            Token.Name.Function.Magic: self.Function,
            Token.Name.Property: self.Name,
            Token.Name.Label: self.Name,
            Token.Name.Namespace: self.Name,
            Token.Name.Other: self.Name,
            Token.Name.Tag: self.Name,
            Token.Name.Variable: self.Name,
            Token.Name.Variable.Class: self.Name,
            Token.Name.Variable.Global: self.Name,
            Token.Name.Variable.Instance: self.Name,
            Token.Name.Variable.Magic: self.Name,
            Token.String: self.String,
            Token.String.Affix: self.String,
            Token.String.Backtick: self.String,
            Token.String.Char: self.String,
            Token.String.Delimiter: self.String,
            Token.String.Doc: self.String,
            Token.String.Double: self.String,
            Token.String.Escape: self.String,
            Token.String.Heredoc: self.String,
            Token.String.Interpol: self.String,
            Token.String.Other: self.String,
            Token.String.Regex: self.String,
            Token.String.Single: self.String,
            Token.String.Symbol: self.String,
            Token.Number: self.Number,
            Token.Number.Bin: self.Number,
            Token.Number.Float: self.Number,
            Token.Number.Hex: self.Number,
            Token.Number.Integer: self.Number,
            Token.Number.Integer.Long: self.Number,
            Token.Number.Oct: self.Number,
            Token.Comment: self.Comment,
            Token.Comment.Hashbang: self.Comment,
            Token.Comment.Multiline: self.Comment,
            Token.Comment.Preproc: self.Comment,
            Token.Comment.PreprocFile: self.Comment,
            Token.Comment.Single: self.Comment,
            Token.Comment.Special: self.Comment,
            # Operator: "#"
            # Operator.Word: "#"
            Token.Punctuation: self.Punctuation,
            Token.Punctuation.Marker: self.Punctuation,
        }



import tkinter as tk
from tkinter import ttk

from biscuit.common.ui import Entry, Frame


class Item(Frame):
    def __init__(self, master, name="Example", *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        self.name = name
        self.description = None  # TODO add descriptions

        self.bg, self.fg, self.highlightbg, _ = self.base.theme.editors.section.values()
        self.config(padx=10, pady=10, bg=self.bg)

        self.lbl = tk.Label(
            self,
            text=self.name,
            font=self.base.settings.uifont_bold,
            anchor=tk.W,
            bg=self.bg,
            fg=self.fg,
        )
        self.lbl.pack(fill=tk.X, expand=True)

    #     self.bind("<Enter>", self.hoverin)
    #     self.bind("<Leave>", self.hoveroff)

    # def hoverin(self, *_):
    #     self.config(bg=self.highlightbg)
    #     self.lbl.config(bg=self.highlightbg)

    # def hoveroff(self, *_):
    #     self.config(bg=self.bg)
    #     self.lbl.config(bg=self.bg)


class DropdownItem(Item):
    def __init__(
        self,
        master,
        name="Example",
        options=["True", "False"],
        default=0,
        *args,
        **kwargs
    ) -> None:
        super().__init__(master, name, *args, **kwargs)

        self.var = tk.StringVar(self, value=options[default])
        m = ttk.OptionMenu(self, self.var, options[default], *options)
        m.config(width=30)
        m.pack(side=tk.LEFT)

    @property
    def value(self) -> str:
        return self.var.get()


class IntegerItem(Item):
    def __init__(self, master, name="Example", default="0", *args, **kwargs) -> None:
        super().__init__(master, name, *args, **kwargs)
        self.base.register(self.validate)

        self.entry = ttk.Entry(
            self,
            font=self.base.settings.uifont,
            width=30,
            validate="key",
            validatecommand=(self.register(self.validate), "%P"),
        )
        self.entry.insert(0, default)
        self.entry.pack(side=tk.LEFT)

    def validate(self, value) -> None:
        return bool(value.isdigit() or value == "")

    @property
    def value(self) -> str:
        return self.entry.get()


class StringItem(Item):
    def __init__(
        self, master, name="Example", default="placeholder", *args, **kwargs
    ) -> None:
        super().__init__(master, name, *args, **kwargs)

        self.entry = ttk.Entry(self, font=self.base.settings.uifont, width=30)
        self.entry.insert(tk.END, default)
        self.entry.pack(side=tk.LEFT)

    @property
    def value(self) -> str:
        return self.entry.get()


class CheckboxItem(Item):
    def __init__(self, master, name="Example", default=True, *args, **kwargs) -> None:
        super().__init__(master, name, *args, **kwargs)

        self.var = tk.BooleanVar(self, value=default)
        ttk.Checkbutton(self, text=name, variable=self.var, cursor="hand2").pack(
            fill=tk.X, anchor=tk.W
        )

    @property
    def value(self) -> str:
        return self.var.get()


# TODO list item with add to list button for taking list values



"""
Parts of the implementation is based on and are inspired by
Porcupine's LSP plugin https://github.com/Akuli/porcupine/
"""

from __future__ import annotations

import itertools
import typing
from pathlib import Path

import tarts as lsp

from biscuit.common import IO

from .data import *
from .handler import EventHandler
from .utils import *

if typing.TYPE_CHECKING:
    from biscuit.editor import Text

    from . import LanguageServerManager


class LangServerClient:
    """Language Server Client

    This class is used to manage the language server client. It is responsible for sending and receiving messages
    to and from the language server. It also manages the requests made by the user. It also manages the deletion of
    the language server client instance when all the tabs using the language server are closed.
    """

    completion_requests: dict[int, tuple[Text, CompletionRequest]] = {}
    gotodef_requests: dict[int, tuple[Text, str]] = {}
    hover_requests: dict[int, tuple[Text, str]] = {}
    outline_requests: dict[int, Text] = {}
    ref_requests: list[tuple[Text, str]] = []
    rename_requests: dict[int, Text] = {}

    def __init__(self, master: LanguageServerManager, tab: Text, root_dir: str) -> None:
        """Initialize the language server client

        Args:
            master (LanguageServerManager): The master language server manager
            tab (Text): The tab using the language server
            root_dir (str): The root directory of the language server"""

        self.master = master
        self.base = master.base
        self.tab = tab
        self.language = tab.language_alias or tab.language
        self.command = master.langservers.get(
            self.language,
            master.langservers.get(
                self.language.lower(),
                master.langservers.get(
                    tab.language,
                    master.langservers.get(
                        tab.language.lower(),
                        master.langservers.get(tab.language_alias, None),
                    ),
                ),
            ),
        )

        self.root_dir = root_dir
        self._counter = itertools.count()

        # TODO: Make this configurable from settings
        # when all tabs using this lsp are closed, delete the lsp client after deletion_delay milliseconds
        self.deletion_delay = 10000
        self.tabs_opened: set[Text] = set()
        self._count = 0

        self.root_uri = Path(self.root_dir).as_uri()
        self.io = IO(self, self.command, self.root_dir)
        self.io.start()
        self.client = lsp.Client(
            process_id=self.io.p.pid,
            root_uri=self.root_uri,
            workspace_folders=[lsp.WorkspaceFolder(uri=self.root_uri, name="Root")],
            trace="verbose",
        )
        self.handler = EventHandler(self)

    def run_loop(self) -> None:
        """Run the language server client loop"""

        if self.run():
            self.base.after(50, self.run_loop)

    def run(self):
        """Run the language server client"""

        self.io.write(self.client.send())
        r = self.io.read()

        if not r:
            return True
        elif r == b"":
            return False

        try:
            for lsp_event in self.client.recv(r):
                self.handler.process(lsp_event)
        except Exception as e:
            print(e)

        return True

    def open_tab(self, tab: Text) -> None:
        """Send the did_open message to the language server client

        Args:
            tab (Text): The tab that is opened"""

        self.tabs_opened.add(tab)

        if self.client.state == lsp.ClientState.NORMAL:
            self.client.did_open(
                lsp.TextDocumentItem(
                    uri=Path(tab.path).as_uri(),
                    languageId=self.language,
                    text=tab.get_all_text(),
                    version=next(self._counter),
                )
            )

    def close_tab(self, tab: Text) -> None:
        """Send the did_close message to the language server client

        Args:
            tab (Text): The tab that is closed"""

        if not tab in self.tabs_opened:
            return

        self.tabs_opened.remove(tab)
        if self.client.state == lsp.ClientState.NORMAL:
            self.client.did_close(
                lsp.TextDocumentIdentifier(uri=Path(tab.path).as_uri())
            )

        if not self.tabs_opened:

            def delayed_removal(self=self):
                if self.tabs_opened:
                    return
                self.master.kill(self)

            if self.master.kill_thread:
                self.base.after_cancel(self.master.kill_thread)
            self.master.kill_thread = self.base.after(50000, delayed_removal)

    def request_completions(self, tab: Text) -> None:
        """Request completions from the language server

        Args:
            tab (Text): The tab requesting completions"""

        if tab.path is None or self.client.state != lsp.ClientState.NORMAL:
            return

        request = CompletionRequest(next(self._counter), tab.get_cursor_pos())
        req_id = self.client.completion(
            text_document_position=lsp.TextDocumentPosition(
                textDocument=lsp.TextDocumentIdentifier(uri=Path(tab.path).as_uri()),
                position=encode_position(request.cursor),
            ),
            context=lsp.CompletionContext(
                triggerKind=lsp.CompletionTriggerKind.INVOKED
            ),
        )

        self.completion_requests[req_id] = (tab, request)

    def request_hover(self, tab: Text) -> None:
        """Request hover information from the language server

        Args:
            tab (Text): The tab requesting hover information"""

        if tab.path is None or self.client.state != lsp.ClientState.NORMAL:
            return

        request_id = self.client.hover(
            lsp.TextDocumentPosition(
                textDocument=lsp.TextDocumentIdentifier(uri=Path(tab.path).as_uri()),
                position=encode_position(tab.get_mouse_pos()),
            )
        )
        self.hover_requests[request_id] = (tab, tab.get_mouse_pos())

    def request_go_to_definition(self, tab: Text) -> None:
        """Request go to definition from the language server

        Args:
            tab (Text): The tab requesting go to definition"""

        if tab.path is None or self.client.state != lsp.ClientState.NORMAL:
            return

        # very bad hack to ignore mouse and use cursor position
        tab.focus_set()
        pos = tab.get_mouse_pos()
        if pos == "1.0":
            pos = tab.get_cursor_pos()

        request_id = self.client.definition(
            lsp.TextDocumentPosition(
                textDocument=lsp.TextDocumentIdentifier(uri=Path(tab.path).as_uri()),
                position=encode_position(pos),
            )
        )
        self.gotodef_requests[request_id] = (tab, pos)

    def request_references(self, tab: Text) -> None:
        """Request references from the language server

        Args:
            tab (Text): The tab requesting references"""

        if tab.path is None or self.client.state != lsp.ClientState.NORMAL:
            return

        tab.focus_set()
        pos = tab.get_mouse_pos()
        if pos == "1.0":
            pos = tab.get_cursor_pos()

        request_id = self.client.references(
            lsp.TextDocumentPosition(
                textDocument=lsp.TextDocumentIdentifier(uri=Path(tab.path).as_uri()),
                position=encode_position(pos),
            )
        )
        self.ref_requests.append((tab, pos))

    def request_rename(self, tab: Text, new_name: str) -> None:
        """Request rename from the language server

        Args:
            tab (Text): The tab requesting rename
            new_name (str): The new name to be used"""

        if tab.path is None or self.client.state != lsp.ClientState.NORMAL:
            return

        tab.focus_set()
        pos = tab.get_cursor_pos()

        request_id = self.client.rename(
            lsp.TextDocumentPosition(
                textDocument=lsp.TextDocumentIdentifier(uri=Path(tab.path).as_uri()),
                position=encode_position(pos),
            ),
            new_name=new_name,
        )
        self.rename_requests[request_id] = tab

    def request_outline(self, tab: Text) -> None:
        """Request outline from the language server

        Args:
            tab (Text): The tab requesting outline"""

        if tab.path is None or self.client.state != lsp.ClientState.NORMAL:
            return

        request_id = self.client.documentSymbol(
            lsp.TextDocumentIdentifier(uri=Path(tab.path).as_uri()),
        )
        self.outline_requests[request_id] = tab

    def send_change_events(self, tab: Text) -> None:
        """Send the did_change message to the language server client

        Args:
            tab (Text): The tab that has changed"""

        if self.client.state != lsp.ClientState.NORMAL:
            return

        self.client.did_change(
            text_document=lsp.VersionedTextDocumentIdentifier(
                uri=Path(tab.path).as_uri(), version=next(self._counter)
            ),
            content_changes=[
                lsp.TextDocumentContentChangeEvent(
                    # range=lsp.Range(
                    #     start=encode_position(tab.get_begin()),
                    #     end=encode_position(tab.get_end()),
                    # ),
                    text=tab.get_all_text(),
                )
            ],
        )



import tkinter as tk
from tkinter import ttk

from .native import Canvas, Frame


class ScrollableFrame(Frame):
    """Scrollable frame widget
    Items can be added to the frame using the add method.
    Items are added to ScrollableFrame.content using the pack manager."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)

        self.scrollbar = ttk.Scrollbar(self, style="EditorScrollbar")
        self.scrollbar.pack(side=tk.RIGHT, fill=tk.Y)

        self.canvas = Canvas(
            self, yscrollcommand=self.scrollbar.set, **self.base.theme.editors
        )
        self.canvas.configure(highlightthickness=0)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)

        self.scrollbar.config(command=self.canvas.yview)

        self.items = []

        self.content = Frame(self.canvas, **self.base.theme.editors)
        self._content = self.canvas.create_window(
            (0, 0), window=self.content, anchor="nw"
        )

        self.content.bind("<Configure>", self._scroll)
        self.canvas.bind("<Configure>", self._configure_canvas)
        self.canvas.bind("<MouseWheel>", self._on_mousewheel)

    def _scroll(self, _) -> None:
        self.canvas.configure(scrollregion=self.canvas.bbox("all"))

    def _configure_canvas(self, event) -> None:
        canvas_width = event.width
        self.canvas.itemconfig(self._content, width=canvas_width)

    def _on_mousewheel(self, event) -> None:
        self.canvas.yview_scroll(int(-1 * (event.delta / 120)), "units")

    def add(self, content, *args, **kwargs) -> None:
        content.pack(in_=self.content, *args, **kwargs)
        self.items.append(content)



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Frame
from biscuit.common.ui import Text as TextW
from biscuit.common.ui import Toplevel

from .renderer import HoverRenderer

if typing.TYPE_CHECKING:
    from biscuit.language.data import HoverResponse
    from src import App

    from ..text import Text


def codeblock(text: str, language: str) -> str:
    return f"```{language}\n{text}\n```\n\n---\n\n"


class Hover(Toplevel):
    def __init__(self, master: App, bd: int = 2, *args, **kw) -> None:
        super().__init__(master, *args, **kw)
        self.overrideredirect(True)
        self.maxsize(600, 200)
        self.config(bg=self.base.theme.primary_background_highlight, bd=bd)

        container = Frame(self, bg=self.base.theme.border)
        container.pack(fill=tk.BOTH, expand=True)

        self.renderer = HoverRenderer(container)
        self.renderer.pack(fill=tk.BOTH, expand=True)

        self.withdraw()

    def update_position(self, pos: str, tab: Text):
        tab.update_idletasks()

        pos_x, pos_y = tab.winfo_rootx(), tab.winfo_rooty()

        pos = tab.index(pos + " wordstart")
        bbox = tab.bbox(pos)
        if not bbox:
            return

        bbx_x, bbx_y, _, _ = bbox
        self.geometry(
            "+{}+{}".format(pos_x + bbx_x - 1, pos_y + bbx_y - self.winfo_height())
        )

    def show(self, tab: Text, response: HoverResponse) -> None:
        if not response or not (response.text or response.docs):
            self.hide()
            return

        docs = ""
        try:
            if response.text[1].strip():
                docs += codeblock(response.text[1].strip(), tab.language_alias)
        except IndexError:
            pass

        if response.docs and response.docs.strip():
            docs += response.docs

        if not docs or not docs.strip():
            self.hide()
            return

        self.renderer.render_markdown(docs)
        self.update_idletasks()
        self.deiconify()
        self.update()
        self.update_position(response.location, tab)

    def hide(self, *_) -> None:
        self.withdraw()



import threading
import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import WrappingLabel

from ..sidebar_view import SideBarView
from .issues import Issues
from .menu import GitHubMenu
from .prs import PRs


class GitHub(SideBarView):
    """View that displays the GitHub issues and pull requests.

    The GitHub view displays the issues and pull requests from the active GitHub repository.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        self.__actions__ = [(Icons.REFRESH, self.on_directory_change)]
        super().__init__(master, *args, **kwargs)
        self.__icon__ = Icons.GITHUB
        self.name = "GitHub"

        self.menu = GitHubMenu(self, "files")
        self.menu.add_checkable("Open Issues", self.toggle_issues, checked=True)
        self.menu.add_checkable("Pull Requests", self.toggle_prs, checked=True)
        self.add_action(Icons.ELLIPSIS, self.menu.show)

        self.issues_enabled = True
        self.prs_enabled = True

        self.git = self.base.git
        self.issues = Issues(self)
        self.prs = PRs(self)

        self.placeholder = WrappingLabel(
            self,
            text="Open a GitHub repository to see issues & pull requests.",
            font=self.base.settings.uifont,
            **self.base.theme.utils.label,
        )
        self.placeholder.pack(fill=tk.X, side=tk.TOP)

        self.base.bind("<<DirectoryChanged>>", self.on_directory_change, add=True)

    def on_directory_change(self, e) -> None:
        """Event handler for directory change event."""

        if not (self.base.active_directory and self.base.git_found):
            self.show_placeholder(
                "Open a GitHub repository to see issues & pull requests."
            )
            return

        repo = self.git.repo
        remote = repo.get_remote_origin()
        if not remote:
            self.base.notifications.info("No remote found.")
            self.show_placeholder("No remote found.")
            return

        if not "github.com" in remote.url:
            self.base.notifications.info("Remote is not a GitHub repository.")
            self.show_placeholder("Remote is not a GitHub repository.")
            return

        try:
            owner, repo_name = repo.get_owner_and_repo(remote.url)
            self.issues.set_url(owner, repo_name)
            self.prs.set_url(owner, repo_name)
        except Exception as e:
            self.base.notifications.error(f"Failed to fetch remote info.")
            self.base.logger.error(f"Failed to fetch remote info: {e}")
            self.show_placeholder("Failed to fetch remote info.")
            return

        self.show_content()
        threading.Thread(target=self.fetch_issues_and_prs, daemon=True).start()

    def fetch_issues_and_prs(self) -> None:
        """Fetches issues and PRs from the current repository."""

        self.issues.fetch()
        self.prs.fetch()

    def show_placeholder(self, text: str) -> None:
        """Adds a placeholder label to the view."""

        self.issues.pack_forget()
        self.prs.pack_forget()
        self.placeholder.config(text=text)
        self.placeholder.pack(fill=tk.X, side=tk.TOP)

    def show_content(self) -> None:
        """Shows the content of the view."""

        self.placeholder.pack_forget()
        if self.issues_enabled:
            self.issues.pack(fill=tk.BOTH, expand=True)
        if self.prs_enabled:
            self.prs.pack(fill=tk.BOTH, expand=True)

    def toggle_issues(self) -> None:
        """Toggles the visibility of the issues view."""

        if self.issues_enabled:
            self.issues.pack_forget()
        else:
            self.issues.pack(fill=tk.BOTH, expand=True)
        self.issues_enabled = not self.issues_enabled

    def toggle_prs(self) -> None:
        """Toggles the visibility of the PRs view."""

        if self.prs_enabled:
            self.prs.pack_forget()
        else:
            self.prs.pack(fill=tk.BOTH, expand=True)
        self.prs_enabled = not self.prs_enabled



import tkinter as tk

from .labels import Label
from .native import Toplevel


class Bubble(Toplevel):
    """Hover bubble for showing information/tips on hover.

    Args:
        master: The parent widget.
        text: The text to be displayed in the bubble.
        bd: The border width of the bubble."""

    def __init__(self, master, text, bd=1, *args, **kw) -> None:
        super().__init__(master, *args, **kw)
        self.overrideredirect(True)
        self.config(bg=self.base.theme.border)
        self.label = Label(
            self,
            text=text,
            padx=5,
            pady=5,
            font=self.base.settings.uifont,
            **self.base.theme.utils.bubble,
        )
        self.label.pack(padx=bd, pady=bd)
        self.withdraw()

    def get_pos(self, *_) -> str:
        """Get the position of the bubble.
        To be reconfigured in inherited classes."""

        return (
            f"+{self.master.winfo_rootx() + self.master.winfo_width() + 5}"
            + f"+{int(self.master.winfo_rooty() + (self.master.winfo_height() - self.winfo_height())/2)}"
        )

    def change_text(self, text) -> None:
        """Change the text of the bubble"""

        self.label.config(text=text)

    def show(self, *_) -> None:
        """Show the bubble"""

        self.update_idletasks()
        self.geometry(self.get_pos())
        self.deiconify()

    def hide(self, *_) -> None:
        """Hide the bubble"""

        self.withdraw()



from __future__ import annotations

import tkinter as tk
import webbrowser
from tkinter import font

from .labels import Label


class LinkLabel(Label):
    """Label that acts as a link"""

    def __init__(
        self, master, text: str, command=lambda _: None, *args, **kwargs
    ) -> None:
        super().__init__(master, text=text, *args, **kwargs)
        self.bg, self.fg, self.hbg, self.hfg = self.base.theme.utils.linklabel.values()

        # self.underlined = font.Font(self, self.base.settings.uifont)
        # self.underlined.configure(underline=True)

        self.config(
            font=self.base.settings.uifont,
            cursor="hand2",
            bg=self.bg,
            fg=self.fg,
        )
        self.bind("<Enter>", self.on_enter)
        self.bind("<Leave>", self.on_leave)
        self.set_command(command)

    def on_enter(self, _) -> None:
        self.config(fg=self.hfg)

    def on_leave(self, _) -> None:
        self.config(fg=self.fg)

    def set_command(self, command) -> None:
        self.bind("<Button-1>", command)

    def pack(self, *args, **kwargs) -> LinkLabel:
        super().pack(*args, **kwargs)
        return self


class WebLinkLabel(LinkLabel):
    """LinkLabel that opens a web link"""

    def __init__(self, master, text, link, *args, **kwargs) -> None:
        super().__init__(master, text=text, *args, **kwargs)
        self.link = link
        self.set_command(self.open_link)

    def open_link(self, *_) -> None:
        webbrowser.open(self.link)



# Palette: https://github.com/morhetz/gruvbox

import sv_ttk

from .theme import *


class GruvboxDark(Theme):
    name = "gruvbox dark"

    biscuit = "#8fbf7c"
    biscuit_dark = "#3c3836"

    border = "#3c3836"
    disabled = "#737373"

    primary_background = "#282828"
    primary_foreground = "#ebdbb2"
    primary_background_highlight = "#3c3836"
    primary_foreground_highlight = "#fbf1c7"

    secondary_background = "#1d2021"
    secondary_foreground = "#ebdbb2"
    secondary_background_highlight = "#32302f"
    secondary_foreground_highlight = "#fbf1c7"

    # syntax
    Keyword = "#f84b3c"
    Name = "#ebdab4"
    Function = "#8fbf7f"
    String = "#b8ba37"
    Number = "#d28797"
    Comment = {"foreground": "#928375", "font": {"slant": "italic"}}
    Punctuation = "#8fbf7f"

    def __init__(self, *args, **kwds) -> None:
        super().__init__(*args, **kwds)
        sv_ttk.use_dark_theme()

        self.editors.linenumbers.number.foreground = "#928375"
        self.editors.linenumbers.number.highlightforeground = "#928375"

        self.editors.linenumbers.breakpoint.background = "#df463a"
        self.editors.linenumbers.breakpoint.highlightbackground = "#df463a"

        self.editors.diff.not_exist = "#424242"
        self.editors.diff.removed = "#452323"
        self.editors.diff.addition = "#203423"

        self.editors.currentword = "#474747"
        self.editors.found = "#203423"
        self.editors.currentline = self.border
        self.editors.selection = "#264f78"
        self.editors.found = "#623315"
        self.editors.foundcurrent = "#9e86b7"
        self.editors.hovertag = "#202b35"

        self.views.panel.logs.time = "#6a9955"
        self.views.panel.logs.caller = "#569cd6"
        self.views.panel.logs.info = "#b5cea8"
        self.views.panel.logs.warning = "#ce9178"
        self.views.panel.logs.error = "#ce9178"

        self.editors.bracket_colors = ("ffd700", "da70d6", "179fff")
        self.editors.activebracket = "#d79921"
        self.editors.hyperlink = "#4583b6"



import tkinter as tk

from biscuit.api import notifications
from biscuit.common.icons import Icons
from biscuit.common.menu import Dropdown

from ..sidebar_view import SideBarView
from .actions import DebuggerActions
from .callstack import CallStack
from .placeholder import DebugPlaceholder
from .variables import Variables

EMPTY_MESSAGE = "No configurations found"


class Debug(SideBarView):
    """A view that displays the debugger variables and call stack.

    - Each editor can have its own debugger.
    - If the editor has a debugger, the debug view will show the variables and call stack of the debugger.
    - Debugger run controls are displayed in the editor toolbar."""

    def __init__(self, master, *args, **kwargs) -> None:
        self.__actions__ = []
        super().__init__(master, *args, **kwargs)
        self.__icon__ = Icons.DEBUG_ALT
        self.name = "Debug"
        self.running = False
        self.manager = self.base.debugger_manager

        self.top.pack_forget()

        self.dropdown = Dropdown(
            self,
            icon=Icons.PLAY,
            callback=self.set_config,
            iconfg="#87d282",
            iconhfg="#87d282",
            empty_message=EMPTY_MESSAGE,
        )
        self.dropdown.icon_label.bind("<Button-1>", self.run_config)
        self.dropdown.pack(fill=tk.X, padx=10, pady=10)

        self.configs = {}
        self.selected_config = ""
        self.refresh_configs()

        self.actionbar = DebuggerActions(self)
        self.variables = Variables(self)
        self.callstack = CallStack(self)

        self.placeholder = DebugPlaceholder(self)
        self.add_item(self.placeholder)

    def refresh_configs(self):
        self.configs = {
            config["name"]: config for config in self.manager.config_loader.configs
        }
        self.selected_config = list(self.configs.keys())[0] if self.configs else None
        self.dropdown.set_items([config for config in self.configs.keys()])
        self.dropdown.change_text(self.selected_config)

    def set_config(self, name: str):
        self.selected_config = name
        self.run_config()

    def run_config(self, *_):
        if not self.selected_config or self.selected_config == EMPTY_MESSAGE:
            self.base.notifications.error(
                "You need to create a launch configuration to run the debugger.",
                actions=[["Configure launch...", self.open_config_editor]],
            )
            return

        try:
            if self.selected_config:
                config = self.configs[self.selected_config]
                language = config["language"]
                if debugger := self.manager.request_debugger(config["language"]):
                    debugger.launch_config(config)
                else:
                    self.base.notifications.error(
                        f"Debugger for {language} not found.",
                        actions=[
                            ["Configure launch...", self.open_config_editor],
                            ["Search Extensions", self.base.sidebar.show_extensions],
                        ],
                    )
        except Exception as e:
            self.base.notifications.error(
                f"Launch configuration {self.selected_config} failed.",
                actions=[
                    ["Configure launch...", self.open_config_editor],
                    ["Logs", self.base.panel.show_logs],
                ],
            )
            self.base.logger.error(
                f"Launch configuration {self.selected_config} failed: {e}"
            )

    def open_config_editor(self, *_):
        path = (
            self.manager.config_loader.path
            if self.base.active_directory
            else ".biscuit/launch.toml"
        )
        self.base.open_editor(str(path))

    def refresh(self):
        if self.manager.config_loader.config_found or self.running:
            self.show()
        else:
            self.hide()

        self.refresh_configs()

    def set_running(self):
        self.running = True
        self.add_item(
            self.actionbar, fill=tk.Y, anchor=tk.CENTER, before=self.variables
        )

    def set_stopped(self):
        self.running = False
        self.actionbar.pack_forget()

    def set_paused(self):
        self.actionbar.pause_btn.toggle_icon()

    def reset(self):
        self.actionbar.pause_btn.reset_icon()
        self.variables.clear()
        self.callstack.clear()

    def show(self):
        self.placeholder.pack_forget()
        self.add_item(self.variables)
        self.add_item(self.callstack)

    def hide(self):
        self.actionbar.pack_forget()
        self.variables.pack_forget()
        self.callstack.pack_forget()
        self.add_item(self.placeholder)



import tkinter as tk

from biscuit.common.ui import Frame

from .items import CheckboxItem, DropdownItem, IntegerItem, StringItem


class Section(Frame):
    """Section for the settings editor to group the items together.

    - Add items to the section to change the settings
    - Add a title to the section to describe the settings
    """

    def __init__(self, master, title="", *args, **kwargs) -> None:
        """Initialize the section with a title

        Args:
            master (tk.Tk): root window
            title (str, optional): title of the section. Defaults to"""

        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.editors, padx=30)

        self.items = []
        tk.Label(
            self,
            text=title,
            font=("Segoi UI", 22, "bold"),
            anchor=tk.W,
            **self.base.theme.editors.labels
        ).pack(fill=tk.X, expand=True)

    def add_dropdown(
        self, name="Example", options=["True", "False"], default=0
    ) -> None:
        """Add a dropdown item to the section

        Args:
            name (str, optional): name of the dropdown. Defaults to "Example".
            options (list, optional): list of options for the dropdown. Defaults to ["True", "False"].
            default (int, optional): default value of the dropdown. Defaults to 0.
        """

        dropdown = DropdownItem(self, name, options, default)
        dropdown.pack(fill=tk.X, expand=True)
        self.items.append(dropdown)

    def add_stringvalue(self, name="Example", default="placeholder") -> None:
        """Add a string text box item to the section

        Args:
            name (str, optional): name of the string. Defaults to "Example".
            default (str, optional): default value of the string. Defaults to "placeholder".
        """

        string = StringItem(self, name, default)
        string.pack(fill=tk.X, expand=True)
        self.items.append(string)

    def add_intvalue(self, name="Example", default="0") -> None:
        """Add an integer text box item to the section

        Args:
            name (str, optional): name of the integer. Defaults to "Example".
            default (int, optional): default value of the integer. Defaults to "0".
        """

        int = IntegerItem(self, name, default)
        int.pack(fill=tk.X, expand=True)
        self.items.append(int)

    def add_checkbox(self, name="Example", default=True) -> None:
        """Add a checkbox item to the section

        Args:
            name (str, optional): name of the checkbox. Defaults to "Example".
            default (bool, optional): default value of the checkbox. Defaults to True.
        """

        dropdown = CheckboxItem(self, name, default)
        dropdown.pack(fill=tk.X, expand=True)
        self.items.append(dropdown)



import tkinter as tk

from biscuit.common.ui import Toplevel
from biscuit.views.explorer import DirectoryTree


class PathView(Toplevel):
    """PathView is a floating window that displays the directory tree of a path.
    PathView is used with the breadcrumbs to display the directory structure of a file.
    User can navigate the tree and open files in the editor.
    """

    def __init__(self, master, width=150, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.width = round(width * self.base.scale)

        self.tree = DirectoryTree(
            self,
            width=width,
            observe_changes=False,
            itembar=False,
            style="secondary.Treeview",
        )
        self.tree.pack()

        self.config(pady=1, padx=1, bg=self.base.theme.border)
        self.overrideredirect(True)
        self.withdraw()

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.configure_bindings()

    def configure_bindings(self) -> None:
        self.bind("<FocusOut>", self.hide)
        self.bind("<Escape>", self.hide)

    def get_popup_x(self, width) -> int:
        return self.winfo_rootx() + int(self.winfo_width() / 2) - int(width / 2)

    def get_popup_y(self) -> int:
        return self.winfo_rooty()

    def hide(self, *_) -> None:
        self.withdraw()

    def show(self, w: tk.Widget) -> None:
        self.update_idletasks()
        x = w.winfo_rootx()
        y = w.winfo_rooty() + w.winfo_height() - 10

        if not w.path:
            return
        self.tree.change_path(w.path)
        self.geometry(f"+{x}+{y}")
        self.deiconify()
        self.lift()
        self.focus_set()



from .ai import AI
from .control import Inspect
from .debug import Debug
from .explorer import Explorer
from .extensions import Extensions
from .github import GitHub
from .logs import Logs
from .outline import Outline
from .panelview import PanelView
from .problems import Problems
from .search import Search
from .sidebar_item import SideBarViewItem
from .sidebar_item_toolbar import SideBarItemToolBar
from .sidebar_view import SideBarView
from .sourcecontrol import SourceControl
from .terminal import Terminal
from .view import View



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from biscuit import App


class Endpoint:
    """Base endpoint class"""

    def __init__(self, base: App) -> None:
        self.base = base



import tkinter as tk
import tkinter.ttk as ttk

from biscuit.common.icons import Icons
from biscuit.common.ui import ButtonsEntry, Frame

from ..sidebar_view import SideBarView
from .results import Results


class Search(SideBarView):
    """The Search view.

    The Search view allows the user to search for text in the active document.
    - Search for text.
    - Replace text.
    - Search with regular expressions.
    - Search case-sensitive.
    - Search whole words.
    - Clear search results.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        self.__actions__ = ()  # TODO: (Icons.REFRESH,), (Icons.CLEAR_ALL,), (Icons.SHADOW_MINUS,))
        super().__init__(master, *args, **kwargs)
        self.__icon__ = Icons.SEARCH
        self.name = "Search"
        self.searchterm = tk.StringVar(self)

        self.top.pack_forget()

        self.results = Results(self, **self.base.theme.views.sidebar.item)

        self.container = Frame(self, **self.base.theme.views.sidebar)
        self.searchbox = ButtonsEntry(
            self.container,
            hint="Search",
            buttons=(
                (Icons.CASE_SENSITIVE, self.results.search_casesensitive),
                (Icons.WHOLE_WORD, self.results.search_wholeword),
                (Icons.REGEX, self.results.search_regex),
                (Icons.SEARCH, self.results.search),
            ),
        )
        self.searchbox.bind("<Return>", self.results.search)

        self.replacebox = ButtonsEntry(
            self.container,
            hint="Replace",
            buttons=((Icons.REPLACE_ALL, self.results.replace_normal),),
        )

        # TODO add ignore folders & extensions boxes

        self.container.pack(fill=tk.BOTH, padx=10, pady=5)
        self.searchbox.pack(fill=tk.X, anchor=tk.N, pady=2)
        self.replacebox.pack(fill=tk.X, side=tk.LEFT, anchor=tk.N, expand=True)

        self.results.pack(fill=tk.BOTH, expand=True, anchor=tk.N)



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Frame, WrappingLabel

if typing.TYPE_CHECKING:
    from biscuit.editor import Text


class OutlineTreePlaceholder(Frame):
    """Placeholder view for the outline tree.

    The OutlineTreePlaceholder view is displayed when there is no outline information available.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=15, pady=10, **self.base.theme.views.sidebar.item)

        self.label = WrappingLabel(
            self,
            text="No outline information",
            font=self.base.settings.uifont,
            anchor=tk.W,
            **self.base.theme.views.sidebar.item.content,
        )
        self.label.pack(fill=tk.X)

    def show(self, tab: Text) -> None:
        self.label.config(
            text=(
                f"No outline information for {tab.filename}"
                if tab
                else "No outline information"
            )
        )
        self.grid()



import tkinter as tk

from biscuit.common import Icons
from biscuit.common.ui import Frame, IconButton

from .sidebar_item import SideBarViewItem
from .view import View


class SideBarView(View):
    """Abstract class for creating a sidebar view."""

    __actions__ = []

    def __init__(
        self, master, name: str = None, icon=Icons.PREVIEW, *args, **kwargs
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.__icon__ = icon
        self.name = name
        self.__name__ = self.__class__.__name__

        self.pack_propagate(False)
        self.config(width=300, **self.base.theme.views.sidebar)

        self.top = Frame(self, **self.base.theme.views.sidebar)
        self.top.pack(fill=tk.X, padx=(15, 10), pady=7)
        self.top.grid_columnconfigure(0, weight=1)

        self.title = tk.Label(
            self.top,
            text=name.upper() if name else self.__class__.__name__.upper(),
            anchor=tk.W,
            font=self.base.settings.uifont,
            **self.base.theme.views.sidebar.title
        )
        self.title.grid(row=0, column=0, sticky=tk.EW)

        self.column = 1
        if not self.__actions__:
            return

        for i in self.__actions__:
            IconButton(self.top, *i).grid(row=0, column=self.column, sticky=tk.E)
            self.column += 1

    def add_action(self, icon: str, event) -> None:
        IconButton(self.top, icon, event).grid(row=0, column=self.column, sticky=tk.E)
        self.column += 1

    def add_item(self, widget, fill=tk.BOTH, expand=True, *args, **kwargs) -> None:
        widget.pack(fill=fill, expand=expand, *args, **kwargs)

    def remove_item(self, widget) -> None:
        widget.pack_forget()

    def create_item(self, name: str = None):
        item = SideBarViewItem(self, name)
        self.add_item(item)
        return item



from biscuit.common.ui import Frame


class View(Frame):
    """Abstract class for creating a view."""

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.views)



from __future__ import annotations

import tkinter as tk
import typing

import mistune
from mistune import InlineParser
from mistune.plugins.abbr import abbr
from mistune.plugins.def_list import def_list
from mistune.plugins.footnotes import footnotes
from mistune.plugins.formatting import strikethrough
from mistune.plugins.speedup import speedup
from mistune.plugins.table import table
from mistune.plugins.task_lists import task_lists
from mistune.plugins.url import url
from pygments import highlight
from pygments.formatters import html
from pygments.lexers import get_lexer_by_name
from pygments.style import Style
from tkinterweb import HtmlFrame

from biscuit.common.ui import Frame, Scrollbar

if typing.TYPE_CHECKING:
    from ...settings.config import Theme
    from .hover import Hover


class HighlightRenderer(mistune.HTMLRenderer):
    def __init__(self, style, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.style = style
        self.formatter = html.HtmlFormatter(style=self.style)

    def block_code(self, code, info=None):
        try:
            if info:
                lexer = get_lexer_by_name(info, stripall=True)
                return highlight(code, lexer, self.formatter)
        except Exception as e:
            print(e)

        return "<pre><code>" + mistune.escape(code) + "</code></pre>"


class BiscuitStyle(Style):
    def __init__(self, theme: Theme):
        super().__init__()
        self.theme = theme
        self.styles = self.theme.syntax

    def __iter__(self):
        for token, color in self.theme.syntax.items():
            if isinstance(color, dict):
                yield token, {
                    "color": color["foreground"][1:],
                    "bold": False,
                    "italic": False,
                    "underline": False,
                    "bgcolor": None,
                    "border": None,
                }
            else:
                yield token, {
                    "color": color[1:],
                    "bold": False,
                    "italic": False,
                    "underline": False,
                    "bgcolor": None,
                    "border": None,
                }


class HoverRenderer(HtmlFrame):
    def __init__(self, master: Hover, *args, **kwargs) -> None:
        super().__init__(master, messages_enabled=False, *args, **kwargs)
        self.base = master.base

        # NOTE: Causing app to crash
        # self.html.shrink(True)

        self.style = BiscuitStyle(self.base.theme)

        self.renderer = HighlightRenderer(self.style, escape=False)
        self.formatter = self.renderer.formatter
        self.markdown = mistune.Markdown(
            renderer=self.renderer,
            inline=InlineParser(),
            plugins=[
                abbr,
                def_list,
                footnotes,
                strikethrough,
                speedup,
                table,
                task_lists,
                url,
            ],
        )

        t = self.base.theme
        pygments_css = self.formatter.get_style_defs(".highlight")
        self.css = f"""
            {pygments_css}

            HTML {{
                width: fit-content;
                height: fit-content;
            }}

            CODE, PRE {{
                font-family: {self.base.settings.font['family']};
                font-size: {self.base.settings.font['size']}pt;
                background-color: {t.border};
                padding: 2px;
            }}

            .highlight {{
                width: 100%;
            }}

            BODY {{
                background-color: {t.secondary_background};
                color: {t.secondary_foreground};
                align-items: left;
            }}
            img {{
                max-width: 100%;
                height: auto;
            }}

            hr {{
                border: 0;
                border-top: 1px solid {t.border};
                max-width: 100%;
            }}
            li{{
                margin-left:1px;
            }}

            tr, th, td {{
                border: 1px solid {t.border};
            }}

            :link    {{ color: {t.biscuit}; }}
            :visited {{ color: {t.biscuit_dark}; }}
            INPUT, TEXTAREA, SELECT, BUTTON {{ 
                background-color: {t.secondary_background};
                color: {t.secondary_foreground_highlight};
            }}
            INPUT[type="submit"],INPUT[type="button"], INPUT[type="reset"], BUTTON {{
                background-color: {t.primary_background};
                color: {t.primary_foreground};
                color: tcl(::tkhtml::if_disabled {t.primary_background}{t.primary_foreground_highlight});
            }}
            """

    def render_markdown(self, docs) -> None:
        self.load_html(self.markdown(docs))
        self.add_css(self.css)
        self.update_idletasks()



from __future__ import annotations

import platform
import sys
import tkinter as tk
import typing
from dataclasses import dataclass
from textwrap import dedent

import psutil

if typing.TYPE_CHECKING:
    from .. import App

from biscuit import __version__


@dataclass
class SysInfo:
    """Get system information"""

    def __init__(self, base: App) -> None:
        """Initialize the system information class with the base app instance."""

        self.base = base

        self.os = platform.system()
        self.version = platform.version()
        self.release = platform.release()
        self.machine = platform.machine()
        self.processor = platform.processor()
        self.python_version = sys.version
        self.tk_version = tk.TclVersion

    @property
    def is_windows(self) -> bool:
        return self.os == "Windows"

    @property
    def is_linux(self) -> bool:
        return self.os == "Linux"

    @property
    def is_macos(self) -> bool:
        return self.os == "Darwin"

    def get_current_stats(self) -> str:
        """Get current CPU and Memory usage"""

        cpu_percent = psutil.cpu_percent(interval=0)
        memory_percent = psutil.virtual_memory().percent
        # memory = psutil.virtual_memory()
        # total_memory_gb = round(memory.total / (1024**3), 2)
        # used_memory_gb = round((memory.total - memory.available) / (1024**3), 2)
        return f"CPU: {cpu_percent}% | Mem: {memory_percent}%"

    def __str__(self) -> None:
        # ikr this is a mess
        return dedent(
            f"""BISCUIT
----------------------------------------

"Life is short, eat more biscuits." 
    - billiam (2020, colorized, not really)

----------------------------------------

Version: {__version__}
Python Version: {self.python_version}
Tcl/Tk Version: {self.tk_version}
OS: {self.os} {self.release} ({self.version})
Processor: {self.processor}

MIT License
Copyright (c) 2021-24 tomlin7"""
        )



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Frame, Toplevel

if typing.TYPE_CHECKING:
    from biscuit import App
    from biscuit.editor.text import Text


class Rename(Toplevel):
    """Floating window for renaming a symbol in the editor.

    The window is placed right below the cursor, and it's width is
    adjusted to the longest item in the list."""

    def __init__(self, master: App, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(pady=1, padx=1, bg=self.base.theme.border)

        self.tab = None
        self.active = False
        self.withdraw()
        self.overrideredirect(True)

        frame = Frame(self, **self.base.theme.palette)
        frame.pack(fill=tk.BOTH, padx=2, pady=2)

        self.text_variable = tk.StringVar()
        self.entry = tk.Entry(
            frame,
            font=self.base.settings.font,
            relief=tk.FLAT,
            highlightcolor=self.base.theme.biscuit,
            textvariable=self.text_variable,
            **self.base.theme.palette.searchbar,
        )
        self.entry.grid(sticky=tk.EW, padx=5, pady=3)

        self.entry.bind("<Return>", self.enter)
        self.bind("<FocusOut>", self.hide())
        self.bind("<Escape>", self.hide)

    def clear(self) -> None:
        self.text_variable.set("")

    def focus(self) -> None:
        self.entry.focus()

    def get(self):
        return self.text_variable.get()

    def enter(self, *_):
        if self.tab:
            self.base.language_server_manager.request_rename(self.tab, self.get())

        self.hide()

    def show(self, tab: Text):
        self.tab = tab
        self.focus()

        current = tab.get_current_fullword()
        if current is None:
            return

        self.text_variable.set(current)
        self.entry.selection_range(0, tk.END)

        self.refresh_geometry(tab)
        self.deiconify()
        self.active = True

    def hide(self, *_):
        self.withdraw()
        self.active = False

    def refresh_geometry(self, tab: Text):
        self.update_idletasks()
        self.geometry("+{}+{}".format(*tab.cursor_wordstart_screen_location()))



from __future__ import annotations

import pprint
import re
import typing

import tarts as lsp

from .data import *
from .utils import *

if typing.TYPE_CHECKING:
    from .client import LangServerClient


class EventHandler:
    """Event Handler

    This class is used to handle the events received from the language server. It is responsible for processing the
    events and updating the UI accordingly.
    """

    def __init__(self, master: LangServerClient):
        self.master = master
        self.client = master.client
        self.base = master.base

    def process(self, e: lsp.Event) -> None:
        """Process the event"""

        # print(e.__class__.__name__.upper(), e, end="\n\n")

        if isinstance(e, lsp.Shutdown):
            self.client.exit()
            return

        if isinstance(e, lsp.ConfigurationRequest):
            self.base.logger.warning(f"Config asked for: {e.items}")
            e.reply([])
            return

        # if isinstance(e, lsp.WorkspaceFolders):
        #     e.reply([lsp.WorkspaceFolder(uri=self.root_uri, name="Root")])
        #     return

        if isinstance(e, lsp.LogMessage):
            self.base.logger.rawlog(e.message, e.type)
            return

        if isinstance(e, lsp.ShowMessage):
            self.base.notifications.notify(e.message, e.type)
            return

        if isinstance(e, lsp.Initialized):
            self.base.logger.info("Capabilities " + pprint.pformat(e.capabilities))
            for tab in self.master.tabs_opened:
                self.master.open_tab(tab)
                # self.master.request_outline(tab)

            self.client.did_change_configuration([])

            self.base.statusbar.process_indicator.hide()
            return

        if isinstance(e, lsp.Completion):
            tab, req = self.master.completion_requests.pop(e.message_id)
            if tab not in self.master.tabs_opened:
                return

            before_cursor = tab.get(f"{req.cursor} linestart", req.cursor)
            match = re.fullmatch(r".*?(\w*)", before_cursor)
            start = len(match.group(1))

            tab.lsp_show_autocomplete(
                Completions(
                    id=req.id,
                    completions=[
                        Completion(
                            item.kind,
                            item.label,
                            tab.index(f"{req.cursor} - {start} chars"),
                            req.cursor,
                            item.insertText or item.label,
                            (item.filterText or item.insertText or item.label)[start:],
                            documentation=get_completion_item_doc(item),
                            # TODO: add documentation
                        )
                        for item in sorted(
                            e.completion_list.items,
                            key=(lambda item: item.sortText or item.label),
                        )
                    ],
                ),
            )
            return

        if isinstance(e, lsp.PublishDiagnostics):
            matching_tabs = [
                tab
                for tab in self.master.tabs_opened
                if tab.path is not None and Path(tab.path).as_uri() == e.uri
            ]
            if not matching_tabs:
                return
            tab = matching_tabs[0]

            tab.lsp_diagnostics(
                [
                    Diagnostic(
                        start=decode_position(diagnostic.range.start),
                        end=decode_position(diagnostic.range.end),
                        message=f"{diagnostic.source}: {diagnostic.message}",
                        severity=(diagnostic.severity),
                    )
                    for diagnostic in e.diagnostics
                ],
            )
            return

        if isinstance(e, lsp.Definition):
            tab, pos = self.master.gotodef_requests.pop(e.message_id)

            tab.lsp_goto_definition(
                Jump(
                    pos=pos,
                    locations=[
                        JumpLocationRange(
                            file_path=str(path),
                            start=decode_position(range.start),
                            end=decode_position(range.end),
                        )
                        for path, range in jump_paths_and_ranges(e.result)
                    ],
                ),
            )
            return

        if isinstance(e, lsp.References):
            if not self.master.ref_requests:
                return

            try:
                tab, pos = self.master.ref_requests.pop(0)
            except Exception as e:
                print(e)

            tab.lsp_goto_definition(
                Jump(
                    pos=pos,
                    locations=[
                        JumpLocationRange(
                            file_path=decode_path_uri(loc.uri),
                            start=decode_position(loc.range.start),
                            end=decode_position(loc.range.end),
                        )
                        for loc in e.result
                    ],
                ),
            )
            return

        if isinstance(e, lsp.WorkspaceEdit):
            tab = self.master.rename_requests.pop(e.message_id)
            if not e.documentChanges:
                return

            tab.lsp_rename(
                WorkspaceEdits(
                    [
                        WorkspaceEdit(
                            file_path=decode_path_uri(i.textDocument.uri),
                            edits=[
                                TextEdit(
                                    start=decode_position(j.range.start),
                                    end=decode_position(j.range.end),
                                    new_text=j.newText,
                                )
                                for j in i.edits
                            ],
                        )
                        for i in e.documentChanges
                    ]
                )
            )

        if isinstance(e, lsp.Hover):
            requesting_tab, location = self.master.hover_requests.pop(e.message_id)
            requesting_tab.lsp_hover(HoverResponse(location, *hover_filter(e.contents)))
            return

        if isinstance(e, lsp.MDocumentSymbols):
            tab = self.master.outline_requests.pop(e.message_id)
            if tab not in self.master.tabs_opened:
                return

            self.base.language_server_manager._update_symbols(tab, e.result)
            self.base.outline.update_symbols(
                tab,
                (
                    e.result
                    if e.result and isinstance(e.result[0], lsp.DocumentSymbol)
                    else to_document_symbol(e.result)
                ),
            )
            return

        # TODO hooks
        if isinstance(e, lsp.ResponseError):
            self.base.logger.error(str(e))
            return

        # DEBUG ones that are not implemented yet
        self.base.logger.trace(e.__class__.__name__.upper())
        # print(e)



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.ui import Frame
from biscuit.views import *

from .panelbar import PanelBar

if typing.TYPE_CHECKING:
    from ..content import Content


class Panel(Frame):
    """Panel
    Panel is a container for panel views. It contains a panelbar and a tabbed view of panel views.
    """

    def __init__(self, master: Content, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master: Content = master

        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)

        self.grid_propagate(False)
        self.config(height=300, bg=self.base.theme.border)

        self.panelbar = PanelBar(self)
        self.panelbar.grid(row=0, column=0, sticky=tk.EW, pady=(0, 1))

        self.views: list[PanelView] = []

        self.default_views = [Problems(self), Logs(self), Inspect(self), Terminal(self)]
        self.add_views(self.default_views)

    def add_views(self, views: list[PanelView]) -> None:
        """Add multiple views to the panel."""

        for view in views:
            self.add_view(view)

    def add_view(self, view: PanelView) -> None:
        """Add a view to the panel.

        Args:
            view (PanelView): view to be added to the panel"""

        self.views.append(view)
        view.generate_actions(self.panelbar)

        self.panelbar.add_tab(view)

    def delete_all_views(self) -> None:
        """Permanently delete all views."""

        for view in self.views:
            view.destroy()

        self.views.clear()

    def delete_view(self, view: PanelView) -> None:
        """Permanently delete a view."""

        view.destroy()
        self.views.remove(view)

    def set_active_view(self, view: PanelView) -> None:
        for tab in self.panelbar.active_tabs:
            if tab.view == view:
                self.panelbar.set_active_tab(tab)
                tab.select()

    @property
    def problems(self) -> Problems:
        return self.default_views[0]

    @property
    def logger(self) -> Logs:
        return self.default_views[1]

    @property
    def control(self) -> Inspect:
        return self.default_views[2]

    @property
    def terminals(self) -> Terminal:
        return self.default_views[3]

    def show_terminal(self, *_) -> None:
        """Shows the terminal if its hidden/minimized."""

        if not self.terminals.active_terminal:
            self.terminals.add_default_terminal()
        self.set_active_view(self.terminals)
        self.show_panel()

    def show_logs(self, *_) -> None:
        """Shows the logs if its hidden/minimized."""

        self.set_active_view(self.logger)
        self.show_panel()

    def toggle_panel(self, *_) -> None:
        """Toggles the panel visibility."""

        self.master.toggle_panel()

    def switch_to_terminal(self, *_) -> None:
        if not self.terminals.active_terminal:
            self.terminals.add_default_terminal()
        self.set_active_view(self.terminals)

    def show_panel(self) -> None:
        """Calls toplevel method to show the panel."""

        self.master.show_panel()



import os
from pathlib import Path

import click
from click.parser import split_opt
from click.utils import make_str

from biscuit import __version__, get_app_instance

from . import editor, extensions, git  # reason: see at bottom


class BiscuitCLI(click.Group):
    path = ""

    def main(self, *args, **kwargs):
        try:
            return super(BiscuitCLI, self).main(*args, **kwargs)
        except Exception:
            path = self.path

            if path:
                path = Path(self.path).resolve()
                path.mkdir(exist_ok=True)

            appdir = Path(__file__).resolve().parent
            app = get_app_instance(appdir, open_path=str(path))
            app.run()

    def resolve_command(self, ctx, args):
        cmd_name = make_str(args[0])
        original_cmd_name = cmd_name

        cmd = self.get_command(ctx, cmd_name)
        if cmd is None and ctx.token_normalize_func is not None:
            cmd_name = ctx.token_normalize_func(cmd_name)
            cmd = self.get_command(ctx, cmd_name)

        if cmd is None and not ctx.resilient_parsing:
            if split_opt(cmd_name)[0]:
                self.parse_args(ctx, ctx.args)

            # ctx.fail(_("No such command {name!r}.").format(name=original_cmd_name))
            self.path = original_cmd_name

        return cmd_name if cmd else None, cmd, args[1:]


@click.group(cls=BiscuitCLI, invoke_without_command=True)
@click.version_option(__version__, "-v", "--version", message="Biscuit v%(version)s")
@click.help_option("-h", "--help")
@click.option("--dev", is_flag=True, help="Run in development mode")
def cli(path=None, dev=False):
    """Biscuit CLI"""

    click.echo(f"Biscuit v{__version__} {'(dev) 🚧' if dev else '🚀'}")


@cli.result_callback()
def process_commands(processors, path=None, dev=False):
    if path:
        path = str(Path(path).resolve())
        click.echo(f"Opening {path}")

    appdir = Path(os.path.abspath(__file__)).parent
    app = get_app_instance(appdir, open_path=path)

    if processors:
        if isinstance(processors, list):
            for processor in processors:
                processor(app)
        else:
            processors(app)

    app.run()


@cli.command("doc")
def docs():
    """Open biscuit documentation

    This command will open the biscuit documentation in the default browser.

    Example:
        biscuit doc
    """

    click.launch("https://tomlin7.github.io/biscuit/")
    exit()


def setup():
    """Setup the CLI commands

    Loads all the commands from the cli extensions and registers them"""

    extensions.register(cli)
    git.register(cli)
    editor.register(cli)


def run():
    """Setup the CLI and run the CLI"""

    setup()
    cli()


if __name__ == "__main__":
    run()



from .endpoint import Endpoint


class Logger(Endpoint):
    """Logger endpoint

    This class is used to interact with the logger API.
    """

    def __init__(self, *a) -> None:
        super().__init__(*a)
        self.logger = self.base.logger

        self.log = self.info = self.show = self.logger.info
        self.warn = self.warning = self.logger.warning
        self.error = self.logger.error



import tkinter as tk


class Separator(tk.Label):
    """A separator for the menu"""

    def __init__(self, master, length=20, *args, **kwargs) -> None:
        """Create a separator

        Args:
            master: The parent widget
            length: The length of the separator
        """

        super().__init__(master, *args, **kwargs)
        self.base = master.base
        self.master = master

        self.config(
            text="—" * round((length * self.base.scale)),
            pady=0,
            height=1,
            **self.base.theme.menu,
            fg=self.base.theme.border
        )



from __future__ import annotations

import tkinter as tk
import typing

from biscuit.common.icons import Icons
from biscuit.common.ui import Frame, IconButton

from .tab import Tab

if typing.TYPE_CHECKING:
    from biscuit.common.ui import IconButton
    from biscuit.layout.content import Content
    from biscuit.views.panelview import PanelView

    from .panel import Panel


class PanelBar(Frame):
    """Panelbar

    Contains the Tabs of panel views and control buttons of each view
    """

    def __init__(self, master: Panel, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        content: Content = master.master

        self.config(**self.base.theme.layout.content.editors.bar)

        self.tab_container = Frame(self, bg=self.base.theme.border)
        self.tab_container.pack(fill=tk.Y, side=tk.LEFT)

        self.active_tabs: list[Tab] = []
        self.active_tab = None

        self.actions: list[IconButton] = []

        # These buttons are common for all panel views
        self.default_actions = (
            (Icons.CLOSE, content.toggle_panel),
            (Icons.CHEVRON_UP, content.toggle_max_panel, Icons.CHEVRON_DOWN),
        )

        for button in self.default_actions:
            IconButton(
                self, *button, **self.base.theme.layout.content.editors.bar.tab
            ).pack(side=tk.RIGHT)

    def add_actions(self, buttons: list[IconButton]) -> None:
        """Add the buttons to the control buttons

        Args:
            buttons (list[IconButton]): buttons to be added"""

        for button in buttons:
            button.pack(side=tk.RIGHT)
            self.actions.append(button)

    def replace_actions(self, buttons: list[IconButton]) -> None:
        """Replace the control buttons with the new buttons

        Args:
            buttons (list[IconButton]): new buttons"""

        self.clear()
        self.add_actions(buttons)

    def clear(self) -> None:
        for button in self.actions:
            button.pack_forget()
        self.actions.clear()

    def add_tab(self, view: PanelView) -> None:
        """Add the view to the tabs

        Args:
            view (PanelView): panel view to be added to the tabs"""

        tab = Tab(self, view)
        tab.pack(fill=tk.Y, side=tk.LEFT, padx=(0, 1), in_=self.tab_container)
        self.active_tabs.append(tab)

        tab.select()

    def set_active_tab(self, selected_tab: Tab) -> None:
        """Set the selected tab as active tab

        Args:
            selected_tab (Tab): selected tab"""

        self.active_tab = selected_tab
        self.replace_actions(selected_tab.view.__actions__)
        for tab in self.active_tabs:
            if tab != selected_tab:
                tab.deselect()



import os

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer


class ExtensionsWatcher(FileSystemEventHandler):
    """Watchdog event handler for the extensions directory.

    The ExtensionsWatcher is used to watch for changes in the extensions directory."""

    def __init__(self, master) -> None:
        self.master = master
        self.base = master.base

        self.observer = Observer()
        self.observer.start()

    def watch(self) -> None:
        if not (self.base.extensionsdir and os.path.isdir(self.base.extensionsdir)):
            try:
                os.makedirs(self.base.extensionsdir, exist_ok=True)
            except:
                self.base.logger.error(
                    "Extensions refreshing failed: no permission to write"
                )
                self.base.notifications.error("Extensions failed: see logs")
                return
        self.observer.schedule(self, self.base.extensionsdir, recursive=True)

    def stop_watch(self) -> None:
        self.observer.stop()

    def on_created(self, *_) -> None:
        self.master.run_fetch_list()

    def on_deleted(self, *_) -> None:
        self.master.run_fetch_list()



from sortedcontainers import SortedDict

from .textindex import TextIndex


class MinClosestKeyDict(SortedDict):
    """A dictionary that returns the value of the closest key that is less than the requested key
    if the requested key is not found in the dictionary."""

    def __setitem__(self, key, value):
        super().__setitem__(TextIndex(key), value)

    def __getitem__(self, key):
        text_key = TextIndex(key)
        if text_key in self:
            return super().__getitem__(text_key), text_key

        # key not found, return the closest key that is less than the requested key
        less = self.bisect_left(text_key)
        if less == 0:
            raise KeyError(f"No key found that is less than {key}")

        closest_key = self.keys()[less - 1]
        return super().__getitem__(closest_key), closest_key


# class MinClosestKeyDict(dict):
#     """A dictionary that returns the value of the closest key that is less than the requested key
#     if the requested key is not found in the dictionary."""

#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)

#         self._sorted_keys = sorted(self.keys())

#     def __setitem__(self, key, value):
#         if key not in self:
#             bisect.insort(self._sorted_keys, key, key=lambda x: x.replace(".", ""))

#         super().__setitem__(key, value)

#     def __delitem__(self, key):
#         super().__delitem__(key)
#         self._sorted_keys.remove(key)

#     def __getitem__(self, key):
#         if key in self:
#             return super().__getitem__(key)

#         pos = bisect.bisect_left(self._sorted_keys, key.replace(".", ""))
#         if pos == 0:
#             raise KeyError(f"No key found that is less than {key}")

#         closest_key = self._sorted_keys[pos - 1]

#         return super().__getitem__(closest_key)



from __future__ import annotations

import tkinter as tk
import typing

from tkinterweb import HtmlFrame

from biscuit.common.ui import Frame, Scrollbar

if typing.TYPE_CHECKING:
    from ..text import TextEditor


class HTMLRenderer(Frame):
    def __init__(self, master, editor: TextEditor, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.editor = editor
        self.config(bg=self.base.theme.border)

        self.text = HtmlFrame(self, messages_enabled=False, vertical_scrollbar=False)
        self.scrollbar = Scrollbar(
            self, orient=tk.VERTICAL, command=self.text.yview, style="EditorScrollbar"
        )

        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)
        self.text.grid(row=0, column=1, sticky=tk.NSEW)
        self.scrollbar.grid(row=0, column=3, sticky=tk.NS)

    def refresh(self, *_):
        htmlcontent = self.editor.text.get_all_text()
        if "<title>" in htmlcontent and "</title>" not in htmlcontent:
            return

        self.text.load_html(htmlcontent)
        t = self.base.theme
        self.text.add_css(
            f"""
            CODE, PRE {{
                font-family: {self.base.settings.font['family']};
                font-size: {self.base.settings.font['size']}pt;
                background-color: {t.border};
                padding: 2px;
            }}
            BODY {{
                background-color: {t.secondary_background};
                color: {t.secondary_foreground};
            }}
            img {{
                max-width: 100%;
                height: auto;
            }}

            hr {{
                border: 0;
                border-top: 1px solid {t.border};
                max-width: 100%;
            }}
            li{{
                margin-left:1px;
            }}
            :link    {{ color: {t.biscuit}; }}
            :visited {{ color: {t.biscuit_dark}; }}
            INPUT, TEXTAREA, SELECT, BUTTON {{ 
                background-color: {t.secondary_background};
                color: {t.secondary_foreground_highlight};
            }}
            INPUT[type="submit"],INPUT[type="button"], INPUT[type="reset"], BUTTON {{
                background-color: {t.primary_background};
                color: {t.primary_foreground};
                color: tcl(::tkhtml::if_disabled {t.primary_background}{t.primary_foreground_highlight});
            }}
            """
        )



import difflib
import os
import re
import threading
import tkinter as tk

from ..editor.editorbase import BaseEditor
from ..editor.text import TextEditor


class DiffPane(TextEditor):
    """Diff Pane

    The diff is shown in two panes, LHS shows last commit and RHS shows the current state of the file.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, minimalist=True, *args, **kwargs)


class DiffEditor(BaseEditor):
    """Diff Editor

    This class is used to show the diff of a file. It uses the `DiffPane` class to show the diff.
    Standard difflib is used to get the diff of the file. The diff is shown in two panes,
    LHS shows last commit and RHS shows the current state of the file."""

    def __init__(
        self,
        master,
        path: str,
        kind: int = None,
        path2: str = None,
        standalone: bool = False,
        language="",
        *args,
        **kwargs,
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(bg=self.base.theme.border)
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.path = path
        self.kind = kind
        self.path2 = path2
        self.editable = True
        self.standalone = standalone

        self.lhs_data = []
        self.rhs_data = []

        self.lhs_last_line = 0
        self.rhs_last_line = 0

        self.lhs = DiffPane(self, path, exists=False)
        self.lhs.grid(row=0, column=0, sticky=tk.NSEW, padx=(0, 1))

        self.rhs = DiffPane(self, path, exists=False)
        self.rhs.grid(row=0, column=1, sticky=tk.NSEW)

        self.differ = difflib.Differ()
        self.debugger = self.rhs.debugger

        self.left = self.lhs.text
        self.right = self.text = self.rhs.text

        self.lhs.scrollbar["command"] = self.on_scrollbar
        self.rhs.scrollbar["command"] = self.on_scrollbar
        self.left["yscrollcommand"] = self.on_textscroll
        self.right["yscrollcommand"] = self.on_textscroll

        self.stipple = self.base.settings.resources.stipple

        self.left.tag_config(
            "addition",
            background=self.base.theme.editors.diff.not_exist,
            bgstipple=f"@{self.stipple}",
        )
        self.left.tag_config("removal", background=self.base.theme.editors.diff.removed)
        self.left.tag_config("removedword", background="red")

        self.right.tag_config(
            "addition", background=self.base.theme.editors.diff.addition
        )
        self.right.tag_config(
            "removal",
            background=self.base.theme.editors.diff.not_exist,
            bgstipple=f"@{self.stipple}",
        )
        self.right.tag_config("addedword", background="green")

        if path2:
            self.run_show_diff()
        else:
            self.run_show_git_diff()

    def on_scrollbar(self, *args) -> None:
        self.left.yview(*args)
        self.lhs.on_scroll()

        self.right.yview(*args)
        self.rhs.on_scroll()

    def on_textscroll(self, *args) -> None:
        self.lhs.scrollbar.set(*args)
        self.rhs.scrollbar.set(*args)
        self.on_scrollbar("moveto", args[0])

    def run_show_diff(self) -> None:
        """Run the show_diff function in a separate thread"""

        threading.Thread(target=self.show_diff).start()

    def show_diff(self) -> None:
        """Show the diff of the file"""

        with open(
            (
                self.path
                if self.standalone
                else os.path.join(self.base.active_directory, self.path)
            ),
            "r",
        ) as f:
            lhs_data = f.read()
        with open(
            (
                self.path2
                if self.standalone
                else os.path.join(self.base.active_directory, self.path2)
            ),
            "r",
        ) as f:
            rhs_data = f.read()

        lhs_lines = [line + "\n" for line in lhs_data.split("\n")]
        rhs_lines = [line + "\n" for line in rhs_data.split("\n")]

        self.diff = list(self.differ.compare(lhs_lines, rhs_lines))
        for i, line in enumerate(self.diff):
            marker = line[0]
            content = line[2:]

            match marker:
                case " ":
                    # line is same in both
                    self.left.write(content)
                    self.right.write(content)

                case "-":
                    # line is only on the left
                    self.lhs_last_line = int(float(self.left.index(tk.INSERT)))
                    self.left.write(content, "removal")

                    # TODO this check is done to make sure if this is a line with modifications
                    # and not a newly added line, but this is not done right.
                    self.right.insert_newline("removal")

                case "+":
                    # line is only on the right
                    self.rhs_last_line = int(float(self.right.index(tk.INSERT)))
                    self.right.write(content, "addition")

                    # TODO this check is done to make sure if this is a line with modifications
                    # and not a newly added line, but this is not done right.
                    self.left.insert_newline("addition")

    def run_show_git_diff(self) -> None:
        """Run the show_diff_git function in a separate thread"""

        threading.Thread(target=self.show_git_diff).start()

    def show_git_diff(self) -> None:
        """Show the diff of the file"""

        try:
            # case: deleted file
            if not self.kind:
                self.left.write(
                    self.base.git.repo.get_commit_filedata(self.path), "removal"
                )
                self.left.update()
                self.left.highlighter.highlight()
                return

            # case: new/untracked file
            if self.kind in (1, 3):
                with open(
                    os.path.join(self.base.active_directory, self.path), "r"
                ) as f:
                    self.right.write(f.read(), "addition")
                    self.right.update()
                    self.right.highlighter.highlight()
                return

            # case: modified file
            lhs_data = self.base.git.repo.get_commit_filedata(self.path)
            with open(os.path.join(self.base.active_directory, self.path), "r") as f:
                rhs_data = f.read()

        except Exception as e:
            self.base.notifications.error(f"Failed to load diff, see logs")
            self.base.logger.error(f"Failed to load diff: {self.path}\n{e}")
            return

        lhs_lines = [line + "\n" for line in lhs_data.split("\n")]
        rhs_lines = [line + "\n" for line in rhs_data.split("\n")]

        self.diff = list(self.differ.compare(lhs_lines, rhs_lines))
        for i, line in enumerate(self.diff):
            marker = line[0]
            content = line[2:]

            match marker:
                case " ":
                    # line is same in both
                    self.left.write(content)
                    self.right.write(content)

                case "-":
                    # line is only on the left
                    self.lhs_last_line = int(float(self.left.index(tk.INSERT)))
                    self.left.write(content, "removal")

                    # TODO this check is done to make sure if this is a line with modifications
                    # and not a newly added line, but this is not done right.
                    self.right.insert_newline("removal")

                case "+":
                    # line is only on the right
                    self.rhs_last_line = int(float(self.right.index(tk.INSERT)))
                    self.right.write(content, "addition")

                    # TODO this check is done to make sure if this is a line with modifications
                    # and not a newly added line, but this is not done right.
                    self.left.insert_newline("addition")

                case "?":
                    # highlight the word changes in the line above
                    if matches := re.finditer(r"\++", content):
                        self.left.delete(
                            str(float(self.rhs_last_line + 1)),
                            str(float(int(float(self.left.index(tk.INSERT))))),
                        )
                        for match in matches:
                            start = f"{self.rhs_last_line}.{match.start()}"
                            end = f"{self.rhs_last_line}.{match.end()}"
                            self.right.tag_add("addedword", start, end)

                    if matches := re.finditer(r"-+", content):
                        self.right.delete(
                            str(float(self.lhs_last_line + 1)),
                            str(float(int(float(self.right.index(tk.INSERT))))),
                        )
                        for match in matches:
                            start = f"{self.lhs_last_line}.{match.start()}"
                            end = f"{self.lhs_last_line}.{match.end()}"
                            self.left.tag_add("removedword", start, end)

            self.left.update()
            self.right.update()

        self.left.highlighter.highlight()
        self.right.highlighter.highlight()

        # Add extra empty lines at the bottom if one side has fewer lines
        lhs_line_count = int(float(self.left.index(tk.END))) - 1
        rhs_line_count = int(float(self.right.index(tk.END))) - 1
        if lhs_line_count > rhs_line_count:
            extra_newlines = lhs_line_count - rhs_line_count
            for _ in range(extra_newlines):
                self.right.insert_newline()
        elif rhs_line_count > lhs_line_count:
            extra_newlines = rhs_line_count - lhs_line_count
            for _ in range(extra_newlines):
                self.left.insert_newline()

        self.left.set_active(False)



from __future__ import annotations

import ast
import bdb
import os
import sys
import threading
import typing
from types import CodeType

from .base import DebuggerBase
from .pyutils import *

if typing.TYPE_CHECKING:
    from .manager import DebuggerManager


class PythonDebugger(DebuggerBase, bdb.Bdb):
    """Python debugger implementation using the built-in `bdb` module.
    Uses biscuit's debugger base class to interact with the app."""

    def __init__(self, manager: DebuggerManager):
        DebuggerBase.__init__(self, manager)
        bdb.Bdb.__init__(self)
        self.current_frame = None
        self.is_running = False
        self.debug = self.base.debug

        # Event to pause the debugger
        self.debugging = threading.Event()

        self.latest_path = ""

    def launch(self, path: str, cwd="") -> None:
        self.reset()
        self.base.sidebar.show_debug()
        self.debug.refresh()
        self.latest_path = path

        script_dir = os.path.dirname(os.path.abspath(path))
        if script_dir not in sys.path:
            sys.path.insert(0, script_dir)

        with open(path, "r") as f:
            code = compile(f.read(), path, "exec")

        self.clear_all_breaks()
        for path, lines in self.breakpoints.items():
            for line in lines:
                self.set_break(path.replace("\\", "/"), line)

        globals_dict = {
            "__name__": "__main__",
            "__file__": path,
            "__builtins__": __builtins__,
        }

        self.is_running = True
        self.thread = threading.Thread(
            target=self._run_debugger,
            name="Python Debugger Thread",
            args=(code, globals_dict, script_dir, cwd),
            daemon=True,
        )
        self.thread.start()

    def _run_debugger(
        self,
        code: CodeType,
        globals_dict: dict[str, typing.Any],
        script_dir: str,
        cwd="",
    ):
        original_cwd = cwd or os.getcwd()
        try:
            os.chdir(script_dir)
            self.set_trace()

            # Update UI
            self.debug.reset()
            self.debug.set_running()

            exec(code, globals_dict)
        except bdb.BdbQuit:
            pass
        except Exception as e:
            print(f"Debugger internal error: {e}")
        finally:
            os.chdir(original_cwd)
            self.is_running = False

            # Update UI
            self.debug.reset()
            self.debug.set_stopped()
            self.debug.refresh()

    def user_line(self, frame):
        self.debug.reset()
        self.current_frame = frame
        if self.break_here(frame):
            self.debugging.clear()

            for d in get_variables(frame):
                self.variables.show(*d)
            self.callstack.show(get_callstack(frame))

            self.debug.set_paused()  # Update UI to show we're paused
            self.debugging.wait()  # Wait for user input

    def user_return(self, frame, return_value): ...

    def user_exception(self, frame, exc_info):
        self.debugging.clear()
        self.current_frame = frame
        self.variables.show(frame.f_locals.items())
        self.callstack.show(get_callstack(frame))
        self.debug.set_paused()
        self.debugging.wait()

    def stop(self):
        self.is_running = False
        self.set_quit()
        self.debugging.set()  # Release any waiting threads

    def restart(self):
        self.stop()
        self.launch(self.latest_path)

    def step_in(self):
        self.set_step()
        self.debugging.set()

    def step_over(self):
        self.set_next(self.current_frame)
        self.debugging.set()

    def step_out(self):
        self.set_return(self.current_frame)
        self.debugging.set()

    def continue_pause(self):
        self.set_continue()

        if self.debugging.is_set():
            self.debugging.clear()
        else:
            self.debugging.set()

    def set_variable(self, name: str, value: str):
        try:
            # evaluate as a literal
            new_value = ast.literal_eval(value)
            print(f"Value of `{name}` changed to {new_value} (literal)")
        except (ValueError, SyntaxError):
            try:
                # evaluate expression in current frame
                new_value = eval(
                    value, self.current_frame.f_globals, self.current_frame.f_locals
                )
                print(f"Value of `{name}` changed to {new_value} (complex expression)")
            except Exception as e:
                print(f"Value of `{name}` cannot be changed: {e}")
                return

        self.current_frame.f_locals[name] = new_value
        save_locals(self.current_frame)

        # Update UI
        self.variables.clear()
        for d in get_variables(self.current_frame):
            self.variables.show(*d)



from __future__ import annotations

import tkinter as tk

import tkinterDnD as dnd

from biscuit.common.ui import Frame, IconButton


class BaseEditor(Frame):
    """Abstract base class for all editor types in Biscuit.

    This class is not meant to be instantiated directly.

    Attributes:
        path (str): The path of the file being edited.
        path2 (str): The path of the file being compared.
        editable (bool): Whether the editor is editable.
        showpath (bool): Whether to show the path of the file.
        content (str): The content of the file being edited.
        diff (bool): Whether the editor is in diff mode.
        run_command_value (str): The command to run the file.
        language (str): The language of the file being edited.
        standalone (bool): Whether the editor is standalone.
        minimalist (bool): Whether the editor is minimalist.
        exists (bool): Whether the file exists.
        debugger (Any): The debugger instance
        runmenu (Any): The run menu instance
        unsupported (bool): Whether the language is unsupported.
        content_hash (str): The hash of the content of the file.
        text (Any): The text widget of the editor.
        __buttons__ (list): A list of buttons to be added to the editor.
        unsaved_changes: Check if there are unsaved changes in the editor.
        breakpoints: Show the breakpoints in the editor.

    Methods:
        ondrop: Handle the drop event on the editor.
        add_button: Add a button to the editor.
        create_buttons: Create the buttons for the editor.
        save: Save the content of the editor."""

    def __init__(
        self, master, path=None, path2=None, editable=True, *args, **kwargs
    ) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(**self.base.theme.editors)

        self.path = path
        self.path2 = path2
        self.editable = editable

        self.showpath = False
        self.content = None
        self.diff = False
        self.run_command_value = ""
        self.debugger = None
        self.language = ""
        self.standalone = False
        self.minimalist = False
        self.exists = True
        self.runmenu = None
        self.unsupported = False
        self.content_hash = ""
        self.text = None
        self.filename = ""

        self.__buttons__ = []

        self.register_drop_target(dnd.FILE)
        self.bind("<<Drop>>", self.ondrop)

    def unsaved_changes(self): ...
    def breakpoints(self, *_): ...
    def save(self, *_): ...

    def ondrop(self, event: tk.Event):
        """Handle the drop event on the editor.

        Args:
            event (tkinter.Event): The event object."""

        if not event.data:
            return

        self.base.open(event.data, warn_for_directory=True)

    def add_button(self, *args):
        self.__buttons__.append(args)

    def create_buttons(self, editorsbar):
        try:
            self.__buttons__ = [
                (
                    IconButton(editorsbar, iconsize=12, hfg_only=True, *button)
                    if isinstance(button, list | tuple)
                    else IconButton(editorsbar, **button)
                )
                for button in self.__buttons__
            ]
        except:
            pass



from __future__ import annotations

import typing

import tarts as lsp

from biscuit.common import ActionSet
from biscuit.common.icons import Icons

from ..sidebar_view import SideBarView
from .outlinetree import OutlineTree

if typing.TYPE_CHECKING:
    from biscuit.editor import Text


class Outline(SideBarView):
    """View that displays the outline of the active document.

    The Outline view displays the outline of the active document.
    - The user can click on a symbol to navigate to it.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        self.__actions__ = [
            #     (Icons.REFRESH,),
            #     (Icons.COLLAPSE_ALL,),
            #     (Icons.ELLIPSIS,),
        ]
        super().__init__(master, *args, **kwargs)
        self.__icon__ = Icons.SYMBOL_CLASS
        self.name = "Outline"

        self.top.pack_forget()

        self.tree = OutlineTree(self)
        self.add_item(self.tree)

    def update_symbols(self, tab: Text, response: list[lsp.DocumentSymbol]) -> str:
        return self.tree.update_symbols(tab, response)

    def get_actionset(self) -> ActionSet: ...



from __future__ import annotations

import typing

from biscuit.common.ui import Icon

from .kinds import kinds

if typing.TYPE_CHECKING:
    from .item import CompletionItem


class Kind(Icon):
    def __init__(self, master: CompletionItem, kind: int = 0, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.master = master
        self.base = master.base

        self.config(**self.base.theme.editors.autocomplete.item, cursor="hand2")
        self.set_kind(kind)

    def set_kind(self, kind: int = 0):
        if not kind:
            return

        kind = kinds[kind - 1] if kind <= len(kinds) and kind > 0 else kinds[0]
        self.set_icon(kind[0])
        self.set_color(kind[1] or self.base.theme.editors.foreground)



import tkinter as tk

from biscuit.common.ui import Button, Frame, ScrollableFrame
from biscuit.editor import BaseEditor

from .searchbar import Searchbar
from .section import Section


class SettingsEditor(BaseEditor):
    """Settings editor for changing the settings of the editor.

    - Add sections for different settings
    - Add items for each section to change the settings
    - Search through the settings to find the desired
    """

    name = "settings"

    def __init__(self, master, exists=False, editable=False, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)
        self.config(padx=100, pady=20, **self.base.theme.editors)
        self.filename = "Settings"

        # TODO searchbar functionality not implemented yet
        # unpack the container and pack a new container for showing results
        self.search = Searchbar(self)
        self.search.pack(fill=tk.X)

        frame = Frame(self, bg=self.base.theme.border)
        frame.pack(fill=tk.BOTH, expand=True)

        self.tree = Frame(frame, **self.base.theme.editors, width=200, pady=20)
        self.tree.pack_propagate(False)
        self.tree.pack(fill=tk.Y, side=tk.LEFT, padx=(0, 1))

        self.sections = []
        self.container = ScrollableFrame(frame)
        self.container.content.config(pady=10)
        self.container.config(**self.base.theme.editors)
        self.container.pack(fill=tk.BOTH, expand=True)

        self.add_sections()

    def add_sections(self):
        self.add_commonly_used()
        self.add_text_editor()

    def add_commonly_used(self):
        """Add commonly used settings to the settings editor"""

        commonly_used = self.add_section(f"Commonly Used")

        commonly_used.add_dropdown("Color Theme", ("dark", "light"))
        commonly_used.add_intvalue("Font Size", 14)
        commonly_used.add_stringvalue("Font Family", "Consolas")
        commonly_used.add_intvalue("Tab Size", 4)

    def add_text_editor(self):
        """Add text editor settings to the settings editor"""

        commonly_used = self.add_section(f"Text Editor")
        commonly_used.add_checkbox("Auto Save", False)
        commonly_used.add_checkbox("Auto Closing Pairs", True)
        commonly_used.add_checkbox("Auto Closing Delete", True)
        commonly_used.add_checkbox("Auto Indent", True)
        commonly_used.add_checkbox("Auto Surround", True)
        commonly_used.add_checkbox("Word Wrap", False)

    def add_section(self, name: str) -> Section:
        """Add a section to the settings editor

        Args:
            name (str): name of the section

        Returns:
            Section: section to add items to"""

        section = Section(self.container.content, name)
        section.pack(fill=tk.X, expand=True)
        self.sections.append(section)

        shortcut = Button(self.tree, name, anchor=tk.W)
        shortcut.pack(fill=tk.X)
        shortcut.config(**self.base.theme.editors.button)

        return section

    def show_result(self, items):
        """Show the search results in the settings editor

        Args:
            items (list): list of items to show in the settings editor"""

        if not any(items):
            return self.show_no_results()

    def show_no_results(self):
        """Show no results found message in the settings editor"""
        ...



import tkinter as tk

from biscuit.common.ui import Tree

from ..sidebar_item import SideBarViewItem


class CallStack(SideBarViewItem):
    """A view that displays the local variables of the debugger.
    Callstack is populated when the debugger is running only."""

    def __init__(self, master, *args, **kwargs) -> None:
        self.title = "Callstack"
        self.__actions__ = ()
        super().__init__(master, itembar=True, *args, **kwargs)

        self.tree = Tree(
            self.content,
            cursor="hand2",
            columns=("fullpath", "line"),
        )
        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.tree.bind("<<TreeviewSelect>>", self.goto_line)

    def show(self, stack: list[tuple[str, str, int]]):
        """Show the call stack from list of tuples (name, filename, line).

        Args:
            stack (list[tuple[str, str, int]]): The call stack to show."""

        self.clear()

        for name, filename, line in stack:
            callstack_str = f"{name} at {filename}, line {line}\n"
            self.tree.add(text=callstack_str, values=(filename, line))

    def clear(self):
        """Clear the call stack."""

        try:
            self.tree.delete(*self.tree.get_children())
        except tk.TclError:
            pass

    def goto_line(self, event):
        """Go to the line of the selected call stack frame.

        Args:
            event (event): The event that triggered the function."""

        item = self.tree.focus()
        if item:
            filename, line = self.tree.item(item, "values")
            self.base.goto_location(filename, str(float(line)))



"""
Code is taken from Porcupine's LSP plugin https://github.com/Akuli/porcupine/

MIT License

Copyright (c) 2017-2023 Akuli

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
from pathlib import Path
from typing import Iterator, Optional
from urllib.request import url2pathname

import tarts as lsp


# Copyright (c) 2021-2024 tomlin7
def to_document_symbol(infos: list[lsp.SymbolInformation]) -> list[lsp.DocumentSymbol]:
    if not infos:
        return []

    infos.sort(
        key=lambda x: (
            (x.location.range.start.line, x.location.range.start.character),
            (-x.location.range.end.line, -x.location.range.end.character),
        )
    )
    res: list[lsp.DocumentSymbol] = []
    parents: list[lsp.DocumentSymbol] = []

    for info in infos:
        element = lsp.DocumentSymbol(
            name=info.name or "Error Symbol",
            kind=info.kind,
            children=[],
            range=info.location.range,
            selectionRange=info.location.range,
        )

        while True:
            if not parents:
                parents.append(element)
                res.append(element)
                break
            parent = parents[-1]
            if contains_range(parent.range, element.range) and not equals_range(
                parent.range, element.range
            ):
                # TODO avoid adding the same named element twice to same parent
                parent.children.append(element)
                parents.append(element)
                break
            parents.pop()

    return res


def get_completion_item_doc(item: lsp.CompletionItem) -> str:
    if not item.documentation:
        return item.label
    if isinstance(item.documentation, lsp.MarkupContent):
        return item.label + "\n" + item.documentation.value
    return item.label + "\n" + item.documentation


def decode_path_uri(file_url: str) -> str:
    if sys.platform == "win32":
        if file_url.startswith("file:///"):
            return url2pathname(file_url[8:])
        else:
            return url2pathname(file_url[5:])
    else:
        return url2pathname(file_url[7:])


def jump_paths_and_ranges(
    locations: list[lsp.Location] | lsp.Location,
) -> Iterator[tuple[Path, lsp.Range]]:
    if not locations:
        locations = []
    if not isinstance(locations, list):
        locations = [locations]

    for location in locations:
        if isinstance(location, lsp.LocationLink):
            yield (decode_path_uri(location.targetUri), location.targetSelectionRange)
        else:
            yield (decode_path_uri(location.uri), location.range)


def hover_filter(
    content: lsp.MarkupContent | str,
) -> list[Optional[list[str, str]], str]:
    if not isinstance(content, lsp.MarkupContent):
        return None, None

    value = content.value.strip()
    if value.startswith("```"):
        value = [i.strip() for i in value[3:].split("```", 1)]
        return value[0].split("\n"), value[1]

    return None, value


def encode_position(pos: str | list[int]) -> lsp.Position:
    if isinstance(pos, str):
        line, column = map(int, pos.split("."))
    else:
        line, column = pos
    return lsp.Position(line=line - 1, character=column)


def decode_position(pos: lsp.Position) -> str:
    return f"{pos.line + 1}.{pos.character}"


def contains_range(range: lsp.Range, pos: lsp.Range) -> bool:
    if pos.start.line < range.start.line or pos.end.line > range.end.line:
        return False
    if (
        pos.start.line == range.start.line
        and pos.start.character < range.start.character
    ):
        return False
    if pos.end.line == range.end.line and pos.end.character > range.end.character:
        return False
    return True


def equals_range(a: lsp.Range, b: lsp.Range) -> bool:
    if not (a or b):
        return True

    return (
        a
        and b
        and a.start.line == b.start.line
        and a.start.character == b.start.character
        and a.end.line == b.end.line
        and a.end.character == b.end.character
    )



from .config import ConfigManager
from .events import EventManager
from .gui import GUIManager


class App(EventManager, GUIManager, ConfigManager):
    """
    BISCUIT CORE
    ------------

    Manages App Configuration, GUI (Tkinter), Events of the App.

    Single point of access to all the important parts of the app. Holds reference to all the components
    of Biscuit and every class of biscuit have a reference back to this `base` class.
    i.e. `self.base` is the reference to this class from any other class of biscuit.

    Usage
    -----

    Example: In order to access the open editor from the git:

    ```py
    class Git:
        def foo(self):
            editor = self.base.editorsmanager.active_editor
            if editor.content and editor.content.exists:
                print(editor.path)
    ```

    Example: Accessing the active editor instance from Foo class of biscuit:

    ```py
    class Foo:
        def foo(self):
            editor = self.base.editorsmanager.active_editor
            if (editor.content and editor.content.exists):
                print(editor.path)
                if (editor.content.editable):
                    self.base.notifications.info(":)")
    ```
    """

    def __init__(self, appdir: str = "", dir: str = "", *args, **kwargs) -> None:
        """Initialize the App.

        Args:
            appdir (str, optional): Directory of the app. Defaults to "".
            dir (str, optional): Directory to open in the editor. Defaults to "".
        """

        super().__init__(*args, **kwargs)
        self.base = self
        self.appdir = appdir

        self.setup()
        self.late_setup()
        self.initialize_app(dir)

    def run(self) -> None:
        """Start the main loop of the app."""

        self.mainloop()
        self.history.dump()
        self.extensions_manager.stop_server()

    def setup(self) -> None:
        """Setup the app."""

        self.initialized = False
        self.setup_path(self.appdir)
        self.setup_configs()
        self.initialize_tk()

    def initialize_app(self, dir: str) -> None:
        """Initialize the editor.

        Args:
            dir (str): Directory to open."""

        self.initialized = True

        self.palette.generate_help_actionset()
        self.logger.info("Initializing editor finished.")

        self.update_idletasks()
        self.menubar.update()
        self.set_title()
        self.open_directory(dir)
        self.focus_force()

    def control_execute(self, text: str) -> None:
        try:
            return eval(text)
        except SyntaxError:
            try:
                exec(text)
            except Exception as e:
                return e
        except Exception as e:
            return e



from __future__ import annotations

import os
import re
import typing
from tkinter import messagebox

from ..common import ActionSet

if typing.TYPE_CHECKING:
    from src import App

git_available = True
try:
    import git

    from .ignore import GitIgnore
    from .repo import GitRepo
except ImportError:
    messagebox.showerror(
        "Git not found",
        "Git is not installed on your PC. Install and add Git to the PATH to use Biscuit",
    )
    git_available = False

URL = re.compile(r"^(?:http)s?://")


# TODO: Many git functions are not implemented yet
class Git(git.Git):
    """Git Integration

    This class is a wrapper around the `git.Git` class to manage git repositories and branches.
    """

    def __init__(self, master: App, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.base = master
        self.repo = None
        self.ignore = GitIgnore(self)
        self.branches = {}

        self.actionset = ActionSet(
            "Manage git branches",
            "branch:",
            self.branches,
            pinned=[
                [
                    "Create new branch: {}",
                    lambda branch=None: self.repo.create_branch(branch),
                ]
            ],
        )

    def late_setup(self) -> None:
        """This function is called after the app is initialized."""

        if not git_available:
            self.base.notifications.warning(
                "Git not found in PATH, git features are disabled"
            )
            self.base.logger.warning("Git not found in PATH, git features are disabled")
            return

        self.base.palette.register_actionset(lambda: self.actionset)

    def check_git(self) -> None:
        """Check if git is available and the active directory is a git repository
        If git is available and the active directory is a git repository, the branches are updated.
        """

        if not (git_available and self.base.active_directory):
            self.base.git_found = False
            return

        try:
            self.repo = GitRepo(self, self.base.active_directory)
            self.base.git_found = True
            self.ignore.load()
            self.base.notifications.info("Git repository found in opened directory")
            self.base.logger.info("Git repository found in opened directory")
            self.update_repo_info()
        except git.exc.InvalidGitRepositoryError:
            self.base.git_found = False

    def update_repo_info(self) -> None:
        """Update the branches in the repository and add them to the actionset
        This function is called when the repository is changed or when the branches are updated.
        """

        if not git_available:
            return

        self.branches = {}

        for branch in self.repo.branches:
            latest_commit = next(self.repo.iter_commits(branch))
            self.branches[branch] = latest_commit.committed_datetime
        self.branches = sorted(self.branches.items(), key=lambda x: x[1], reverse=True)

        # TODO: make use of the commit_time in palette items
        self.actionset.update(
            [
                (str(branch), lambda e=None, b=branch: self.repo.switch_to_branch(b))
                for branch, commit_time in self.branches
            ]
        )

    def get_version(self) -> str:
        if not git_available:
            return

        return self.version()

    @property
    def active_branch(self) -> str:
        if not git_available:
            return

        return self.repo.active_branch

    def checkout(self, branch: str) -> None:
        if not git_available:
            return

        self.repo.index.checkout(branch)
        self.base.notifications.info(f"Checked out branch {branch}")
        self.base.logger.info(f"Checked out branch {branch}")

    def clone(self, url: str, dir: str) -> str:
        """Clone a git repository to the specified directory

        Args:
            url (str): The url of the git repository
            dir (str): The directory to clone the repository to

        Returns:
            str: The path to the cloned repository"""

        if not URL.match(url):
            # assumes github as repo host
            url = f"http://github.com/{url}"

        if name := self.repo_name(url):
            dir = os.path.join(dir, name)
            GitRepo(self).clone_from(url, dir)
            return dir

        raise Exception(f"The url `{url}` does not point to a git repo")

    def repo_name(self, url: str) -> None:
        match = re.search(r"/([^/]+?)(\.git)?$", url)
        if match:
            return match.group(1)



import google.generativeai as ai
from google.generativeai.types import HarmBlockThreshold, HarmCategory

from .model import ChatModelInterface


class Gemini1p5Flash(ChatModelInterface):
    def __init__(self, api_key: str, prompt=""):
        super().__init__(api_key, prompt)
        self.initialize_model()

    def initialize_model(self):
        ai.configure(api_key=self.api_key)
        self.model = ai.GenerativeModel(
            "gemini-1.5-flash",
            safety_settings={
                HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_NONE,
                HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_NONE,
                HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_NONE,
                HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_NONE,
            },
        )

    def set_api_key(self, api_key: str):
        self.api_key = api_key
        self.initialize_model()

    def set_prompt(self, prompt: str):
        self.prompt = prompt

    def start_chat(self):
        self.chat = self.model.start_chat()
        return self.chat

    def send_message(self, message: str):
        return self.chat.send_message([self.prompt, message]).text



from __future__ import annotations

import os
import typing
from ast import Tuple

import git
import git.exc

if typing.TYPE_CHECKING:
    from .git import Git


class GitRepo(git.Repo):
    """Git Repository

    This class is a wrapper around the `git.Repo` class to manage git repositories and branches.
    """

    def __init__(self, master: Git = None, path=None, *args, **kwargs) -> None:
        super().__init__(path, *args, **kwargs)
        self.master = master
        try:
            self.base = master.base
        except AttributeError:
            self.base = master

        self.path = path
        self.config = self.config_reader()

        self.author_name = self.config.get_value("user", "name")
        self.author_email = self.config.get_value("user", "email")
        self.author = git.Actor(self.author_name, self.author_email)

    def switch_to_branch(self, branch: git.Head):
        self.git.checkout(str(branch))
        self.master.update_repo_info()
        self.base.statusbar.update_git_info()
        self.base.explorer.directory.refresh_root()

    def create_branch(self, branch: str):
        if not branch:
            self.base.notifications.error("Branch name cannot be empty")
            return

        self.create_head(branch.strip())
        self.switch_to_branch(branch)

    def get_untracked_files(self) -> list:
        return list(self.untracked_files)

    def get_added_files(self) -> list:
        return [item.a_path for item in self.index.diff(None).iter_change_type("A")]

    def get_deleted_files(self) -> list:
        return [item.a_path for item in self.index.diff(None).iter_change_type("D")]

    def get_modified_files(self) -> list:
        return [item.a_path for item in self.index.diff(None).iter_change_type("M")]

    def get_staged_added_files(self) -> list:
        return [item.a_path for item in self.index.diff("HEAD").iter_change_type("D")]

    def get_staged_deleted_files(self) -> list:
        return [item.a_path for item in self.index.diff("HEAD").iter_change_type("A")]

    def get_staged_modified_files(self) -> list:
        return [item.a_path for item in self.index.diff("HEAD").iter_change_type("M")]

    def get_latest_commit(self):
        return self.head.commit

    def get_commit_filedata(self, filename) -> str:
        return self.head.commit.tree[filename].data_stream.read().decode("utf-8")

    def stage_files(self, *paths: Tuple[str, int]) -> None:
        """Stage files to be committed

        Change types:
        0: Deleted
        1: Added
        2: Modified
        3: Untracked

        Args:
            *paths (Tuple[path, change_type]): A list of paths and their change types"""

        for path, change_type in paths:
            # change type can be      0,       1,     2,        3
            # respectively represents Deleted, Added, Modified, Untracked
            if change_type == 0:
                self.do(self.index.remove, [path])
            else:
                self.do(self.index.add, [path])

    def unstage_files(self, *paths: str) -> None:
        self.index.reset(paths=paths)

    def discard_changes(self, *paths: str) -> None:
        try:
            self.git.checkout("--", *paths)
        except git.exc.GitCommandError as e:
            for path in paths:
                if path in self.get_untracked_files():
                    os.remove(path)

    def commit_files(self, message="", **kwargs):
        if not message:
            message = "Commit changes"

        return self.index.commit(message, author=self.author, **kwargs)

    def push_files(self, remote="origin", branch="", **kwargs):
        if not branch:
            branch = self.active_branch.name
        return self.do(self.remotes[remote].push, branch, **kwargs)

    def pull_files(self, remote="", branch="", **kwargs):
        if not remote:
            remote = "origin"
        if not branch:
            branch = self.active_branch.name
        return self.do(self.remotes[remote].pull, branch, **kwargs)

    def do(self, fn, *args, **kwargs):
        # error handling
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            self.base.notifications.error(e)
            self.base.logger.error(e)

    def get_remote(self, name):
        return self.remotes[name]

    def get_remote_origin(self):
        return self.get_remote("origin")

    def get_remote_url(self, name):
        return self.strip_github_url(self.get_remote(name).url)

    def get_remote_origin_url(self):
        return self.get_remote_url("origin")

    def strip_github_url(self, url):
        if url.endswith(".git"):
            url = url[:-4]
        return url

    def get_owner_and_repo(self, url):
        url = self.strip_github_url(url).split("/")
        if len(url) == 2:
            # SSH url (git@github.com:username/repo.git)
            return url[0].split(":")[1], url[1]
        return url[-2], url[-1]

    # TODO: push, pull, fetch
    def get_active_branch(self):
        return self.active_branch.name



__author__ = "nfoert"

import os
import re
import tkinter as tk
from tkinter.messagebox import askyesno

from biscuit.common.ui import Frame, Label, Tree


class Results(Frame):
    """The Results view.

    The Results view displays the results of a search.
    - Show search results.
    - Replace text in search results.
    - Search case-sensitive.
    - Search whole words.
    - Search with regular expressions.
    - Clear search results.
    """

    def __init__(self, master, *args, **kwargs) -> None:
        super().__init__(master, *args, **kwargs)

        self.label = Label(self, text="Search")
        self.label.pack(fill=tk.X)

        self.treeview = Tree(self)
        self.treeview.pack(fill=tk.BOTH, expand=True)

        self.ignore_folders = [
            ".git",
            "__pycache__",
            "venv",
            "node_modules",
            "docs",
            "build",
            "dist",
            "bin",
            "obj",
            "out",
            "target",
        ]
        self.ignore_exts = []

        self.results = []

        self.searching = False
        self.case_sensitive = False
        self.whole_word = False
        self.regex = False

        self.replacing = False
        self.r_matchcase = False

    def add_item(self, parent: str, index: str, text: str, open=False) -> str:
        "Add an item to the tree for the search"
        try:
            result = self.treeview.insert(
                parent=parent, index=index, text=text, open=open
            )
            return result

        except Exception as e:
            print("Failed to add item", e)

    def clear_tree(self) -> None:
        "Clear all items from the tree"
        self.treeview.delete(*self.treeview.get_children())

    def delete_item(self, item: str) -> None:
        "Remove an item from the tree"
        self.treeview.delete(item)

    def click(self, _) -> None:
        "Event run when an item is clicked in the tree"
        # Thanks to tobias_k's answer here
        # https://stackoverflow.com/questions/30614279/tkinter-treeview-get-selected-item-values
        item = self.treeview.focus()

        try:  # Click on child item
            self.base.goto_location(
                self.treeview.item(item)["tags"][0],
                str(float(self.treeview.item(item)["tags"][1])),
            )

        except IndexError:  # Click on parent item
            print("You clicked on a parent item")

    def search_casesensitive(self, _) -> None:
        "Do a case sensitive search"
        self.case_sensitive = True
        self.search()
        self.case_sensitive = False

    def search_wholeword(self, _) -> None:
        "Do a whole word search"
        self.whole_word = True
        self.search()
        self.whole_word = False

    def search_regex(self, _) -> None:
        "Do a regex search"
        self.regex = True
        self.search()
        self.regex = False

    # Thanks to pythontutorial.net/tkinter/tkinter-askyesno
    def replace_normal(self, _) -> None:
        "Do a normal replace"
        self.replace()

    def search(self, *_) -> None:
        """
        Find every file in the selected directory and then search for occurrences in it
        """
        if not self.searching:
            self.searching = True

            self.label.config(text="Searching...")
            self.base.notifications.info("Searching...")

            found_files = []
            search_string = self.master.searchbox.get()

            self.clear_tree()
            self.results = []

            if self.base.active_directory:
                for root, _, files in os.walk(self.base.active_directory):
                    if root in self.ignore_folders:
                        continue

                    for file in files:
                        file_path = os.path.join(root, file)

                        if any(file.endswith(ext) for ext in self.ignore_exts):
                            continue

                        result = self.search_in_file(file_path, search_string)

                        if result:
                            found_files.append(result[0])
                            self.label.config(
                                text=f"Searched {len(found_files)} files..."
                            )

                        self.base.root.update()

                if len(found_files) > 0:
                    self.label.config(
                        text=f"{len(self.results)} results for '{search_string}'"
                    )

                else:
                    self.label.config(text="No results.")

            else:
                self.label.config(text="No folder selected.")

            self.base.notifications.hide()
            self.searching = False

        else:
            self.base.notifications.warning("Already searching!")

    def search_in_file(self, file_path: str, search_string: str) -> list:
        """
        Search a file for occurrences
        - Case insensitive
        - Case sensitive
        - Whole word
        - Regex
        """
        found = False
        occurrences = 0
        result_lines = []

        try:
            with open(file_path, "r", encoding="utf-8") as f:
                for line_number, line in enumerate(f, start=1):
                    if self.case_sensitive:
                        if search_string in line:
                            found = True
                            occurrences += line.count(search_string)
                            result_lines.append((line_number, line.strip()))
                            text = search_string

                    elif self.whole_word:
                        result = re.search(r"\b" + search_string + r"\b", line)
                        if result:
                            found = True
                            occurrences += len(
                                re.findall(r"\b" + search_string + r"\b", line)
                            )
                            result_lines.append((line_number, line.strip()))
                            text = result.group()

                    elif self.regex:
                        result = re.search(search_string, line)

                        if result:
                            found = True
                            occurrences += len(re.findall(search_string, line))
                            result_lines.append((line_number, line.strip()))
                            text = result.group()

                    else:
                        if search_string.lower() in line.lower():
                            found = True
                            occurrences += line.count(search_string.lower())
                            result_lines.append((line_number, line.strip()))
                            text = search_string
        except UnicodeDecodeError:
            # case: file is non-text
            return

        if found:
            parent = self.add_item(
                parent="",
                index=tk.END,
                open=True,
                text=f"{os.path.basename(file_path)} | {file_path}",
            )

            for line_number, line in result_lines:
                child_elm = self.add_item(
                    parent=parent, index=tk.END, text=f"line {line_number}: {line}"
                )
                self.treeview.item(child_elm, tags=(file_path, line_number))
                self.treeview.bind("<Double-1>", self.click)

                self.results.append(
                    {"file_path": file_path, "line": line_number, "text": text}
                )

            return [file_path, occurrences, line_number]

    def replace(self) -> None:
        """
        Replace all occurrences from the search with new text

        TODO: It looks like files aren't refreshed after the replace happens.
        Opening a file which had contents replaced shows the old contents
        """
        replace_string = self.master.replacebox.get()

        if self.searching:
            self.base.notifications.warning("Please wait for search to complete")
        elif self.replacing:
            self.base.notifications.warning("Already replacing!")
        else:
            self.replacing = True

            if not self.r_matchcase:
                if not self.results:
                    self.label.config(text="Nothing to replace!")

                else:
                    answer = askyesno(
                        "Replace Confirmation",
                        f"Are you sure you want to apply a replace to {len(self.results)} occurrences?",
                    )
                    total = len(self.results)
                    so_far = 0

                    if answer:
                        for item in self.results:
                            so_far += 1

                            self.label.config(
                                text=f"Replacing... {round(round(so_far / total, 2) * 100)}%"
                            )
                            self.base.root.update()

                            with open(item["file_path"], "r", encoding="utf-8") as file:
                                data = file.readlines()
                                data[item["line"] - 1] = data[item["line"] - 1].replace(
                                    item["text"], replace_string
                                )

                            with open(item["file_path"], "w", encoding="utf-8") as file:
                                file.writelines(data)

                        self.label.config(text="Done replacing.")
                        self.clear_tree()
                        self.results = []

            else:
                print("Replace with matchcase! (Not implemented yet)")

            self.replacing = False



from __future__ import annotations

import importlib.resources as resources
import os
import sys
import tkinter as tk
import typing
from contextlib import ExitStack
from importlib.resources import as_file, files

if typing.TYPE_CHECKING:
    from .. import App


class Resources:
    """Loads and manages resources for biscuit."""

    def __init__(self, master) -> None:
        self.base: App = master.base
        self.exit_stack = ExitStack()
        self.load_data()

    def load_data(self) -> None:
        self.logo = self.load_image("logo.png")
        self.stipple = self.get_res_path("bitmap/stipple.xbm")
        self.indent_guide = self.get_res_path("bitmap/indentguide.xbm")
        self.line = self.get_res_path("bitmap/line.xbm")

    def get_res_path(self, relative_path: str) -> str:
        if self.base.frozen:
            # PyInstaller creates a temp folder and stores path in _MEIPASS
            return os.path.join(self.base.resdir, relative_path)
        else:
            # Running from source / package
            try:
                path = self.exit_stack.enter_context(
                    as_file(files("resources").joinpath(relative_path))
                )
                return str(path)
            except:
                # Fallback to the old method if resource not found
                return self._fallback_get_res_path(relative_path)

    def _fallback_get_res_path(self, relative_path: str) -> str:
        paths = [
            os.path.join(self.base.resdir, relative_path),
            os.path.join(self.base.fallback_resdir, relative_path),
            os.path.join(self.base.second_fallback_resdir, relative_path),
        ]
        for path in paths:
            if os.path.exists(path):
                return path

        raise FileNotFoundError(f"Resource not found: {relative_path}")

    def get_res_path_platform(self, relative_path: str) -> str:
        platform = "win32" if self.base.system.is_windows else "linux"
        return self.get_res_path(os.path.join(platform, relative_path))

    def get_font(self, path: str) -> str:
        return self.get_res_path(os.path.join("fonts", path))

    def load_image(self, path: str) -> tk.PhotoImage:
        return tk.PhotoImage(file=self.get_res_path(path))

    def __del__(self):
        self.exit_stack.close()



import tkinter as tk

from biscuit.common.icons import Icons
from biscuit.common.ui import IconLabelButton


class Checkable(IconLabelButton):
    """A checkable menu item
    Inherits from IconLabelButton"""

    def __init__(
        self, master, text, command=lambda *_: ..., checked=False, *args, **kwargs
    ) -> None:
        """Create a checkable menu item

        Args:
            master: The parent widget
            text: The text to display on the menu item
            command: The command to run when the item is clicked
            checked: The initial checked state
            *args: Additional arguments to pass to the Icon
        """

        super().__init__(
            master,
            text,
            Icons.CHECK,
            command,
            expandicon=False,
            iconsize=10,
            icon_visible=checked,
            *args,
            **kwargs
        )

        self.command = command
        self.bg, self.fg, self.hbg, self.hfg = self.base.theme.menu.item.values()
        self.on_leave()

    def on_click(self, *_) -> None:
        self.master.hide()
        self.toggle_icon()
        return super().on_click(*_)



# Credits: RINO-GAELICO

from __future__ import annotations

import os
import sqlite3 as sq
import typing

if typing.TYPE_CHECKING:
    from biscuit import App


class SessionManager:
    def __init__(self, base: App):
        self.base = base

        # Initialize the session database connection
        self.base_dir = self.base.datadir
        self.session_db_path = os.path.join(self.base_dir, "session.db")
        self.db = sq.connect(self.session_db_path)
        self.cursor = self.db.cursor()

        # Ensure the session table is created
        self._create_session_table()

    def _create_session_table(self):
        """Create the session table if it doesn't exist."""
        self.cursor.executescript(
            """
            CREATE TABLE IF NOT EXISTS session (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                file_path TEXT,
                folder_path TEXT
            );
            """
        )

    def restore_session(self):
        """Restore the session from the database."""
        opened_files = []
        active_directory = ""

        self.cursor.execute("SELECT * FROM session")
        for row in self.cursor.fetchall():
            if row[1]:
                opened_files.append(row[1])
            elif row[2]:
                active_directory = row[2]

        self.base.open_directory(active_directory)
        self.base.open_files(opened_files)

    def clear_session(self):
        """Clear the session table."""
        self.cursor.execute("DELETE FROM session")

    def save_session(self, opened_files, active_directory):
        """Save the currently opened files and directories into the session."""
        for file_path in opened_files:
            self.cursor.execute(
                "INSERT INTO session (file_path) VALUES (?)", (file_path,)
            )

        self.cursor.execute(
            "INSERT INTO session (folder_path) VALUES (?)", (active_directory,)
        )

        self.db.commit()

    def close(self):
        """Close the database connection."""
        self.db.close()
